@@ -21,19 +21,22 @@
 // 2001-11-02 Paul Kienzle <pkienzle@users.sf.net>
 // * fixed mxGetString to put in the zero-terminator
 
-#include <float.h>
-#include <iomanip.h>
+#include <cfloat>
+#include <iomanip>
+#include <set>
+#include <string>
+typedef void * Pix;
+typedef std::set<Pix> MemSet;
+
 extern "C" {
-#include <stdlib.h>
-#include <setjmp.h>
+#include <cstdlib>
+#include <csetjmp>
   extern const char *mexFunctionName;
 } ;
 
+#include <octave/config.h>
 #include <octave/oct.h>
 #include <octave/pager.h>
-#include <octave/SLList.h>
-// XXX FIXME XXX --- this belongs in SLList.h, not SLList.cc
-template <class T> SLList<T>::~SLList (void) { clear(); }
 #include <octave/f77-fcn.h>
 #include <octave/unwind-prot.h>
 #include <octave/lo-mappers.h>
@@ -41,15 +44,17 @@ template <class T> SLList<T>::~SLList (void) { clear(); }
 #include <octave/parse.h>
 #include <octave/toplev.h>
 #include <octave/variables.h>
+#include <octave/oct-map.h>
+#include <octave/str-vec.h>
 
-// ================ Octave 2.0 support ==================
+// ===== Cruft to support ancient versions of octave ========
 #if defined(HAVE_OCTAVE_20)
 #include <octave/symtab.h>
 class unwind_protect
 {
 public:
-  static void add(cleanup_func fptr, void *ptr) 
-  { 
+  static void add(cleanup_func fptr, void *ptr)
+  {
     add_unwind_protect (fptr, ptr);
   }
   static void run(void)
@@ -59,7 +64,7 @@ public:
 } ;
 
 static octave_value_list
-eval_string (const string& s, int print, int& parse_status, int nargout) 
+eval_string (const string& s, int print, int& parse_status, int nargout)
 {
   return eval_string(s,print,parse_status);
 }
@@ -94,98 +99,30 @@ octave_vformat (std::ostream& os, const char *fmt, va_list args)
   return retval;
 }
 
+#endif /* defined(HAVE_OCTAVE_20) */
+
+#ifndef OCTAVE_LOCAL_BUFFER
+#include <vector>
+#define OCTAVE_LOCAL_BUFFER(T, buf, size) \
+  std::vector<T> buf ## _vector (size); \
+  T *buf = &(buf ## _vector[0])
 #endif
 
+// ============== End cruft support =======================
+
 #if 0
 #define TRACEFN cout << __FUNCTION__ << endl << flush
 #else
 #define TRACEFN do { } while(0)
 #endif
 
-/* ============== mxArray data type ============= */
-// Class mxArray is not much more than a struct for keeping together
-// dimensions and data.  It doesn't even ensure consistency between
-// the dimensions and the data.  Unfortunately you can't do better
-// than this without restricting the operations available in Matlab
-// for directly manipulating its mxArray type.
-
-typedef unsigned short mxChar;
-const int mxMAXNAM=64;
-
-class mxArray {
-public:
-  ~mxArray () { }
-
-  int rows() const { return nr; }
-  int columns() const { return nc; }
-  void rows(int r) { nr = r; }
-  void columns(int c) { nc = c; }
-
-  double *imag() const { return pi; }
-  double *real() const { return pr; }
-  void imag(double *p) { pi = p; }
-  void real(double *p) { pr = p; }
-
-  bool is_empty() const { return nr==0 || nc==0; }
-  bool is_numeric() const { return !isstr && (pr != NULL || nr==0 || nc==0); }
-  bool is_complex() const { return pi != NULL; }
-  bool is_sparse() const { return false; }
-
-  bool is_string() const { return isstr; }
-  void is_string(bool set) { isstr = set; }
-
-  const char* name() const { return aname; }
-  void name(const char *nm) { 
-    strncpy(aname,nm,mxMAXNAM); 
-    aname[mxMAXNAM]='\0'; 
-  }
-
-  octave_value as_octave_value() const;
-
-private:
-  int nr, nc;
-  double *pr, *pi;
-  bool isstr;
-  char aname[mxMAXNAM+1];
-} ;
-
-octave_value mxArray::as_octave_value() const
-{
-  octave_value ret;
-  if (isstr)
-    {
-      charMatrix chm(nr,nc);
-      char *pchm = chm.fortran_vec();
-      for (int i=0; i < nr*nc; i++) 
-	pchm[i] = NINT(pr[i]);
-      ret = octave_value(chm, true);
-    }
-  else if (pi)
-    {
-      ComplexMatrix cm(nr, nc);
-      Complex *pcm = cm.fortran_vec();
-      for (int i=0; i < nr*nc; i++) pcm[i] = Complex(pr[i], pi[i]);
-      ret = cm;
-    }
-  else if (pr)
-    {
-      Matrix m(nr,nc);
-      double *pm = m.fortran_vec();
-      memcpy(pm, pr, nr*nc*sizeof(double));
-      ret = m;
-    }
-  else
-    ret = Matrix(0,0);
-
-  return ret;
-}
-
 /* ========== mex file context ============= */
 // Class mex keeps track of all memory allocated and frees anything
 // not explicitly marked persistent when the it is destroyed.  It also
 // maintains the setjump/longjump buffer required for non-local exit
 // from the mex file, and any other state local to this instance of
 // the mex function invocation.
+class mxArray;
 
 class mex {
 
@@ -219,6 +156,10 @@ public:
   // freed on exit unless marked as persistent
   mxArray *make_value(const octave_value&);
 
+  // make a new structure value and initialize with empty matrices
+  // XXX FIXME XXX does this leak memory?  Is it persistent?
+  mxArray *make_value(int nr, int nc, const string_vector& keys);
+
   // free an array and its contents
   void free_value(mxArray* ptr);
 
@@ -226,18 +167,18 @@ public:
   void persistent(mxArray* ptr);
 
   // 1 if error should be returned to MEX file, 0 if abort
-  int trap_feval_error; 
+  int trap_feval_error;
 
   // longjmp return point if mexErrMsgTxt or error
-  jmp_buf jump;  
-  
+  jmp_buf jump;
+
   // trigger a long jump back to the mex calling function
   void abort() { longjmp(jump, 1); }
 
 private:
 
   // list of memory resources that need to be freed upon exit
-  SLList<Pix> memlist;
+  MemSet memlist;
 
   // mark a pointer to be freed on exit
   void mark(Pix p);
@@ -248,52 +189,39 @@ private:
 
 } ;
 
+// Current context
+mex* __mex = NULL;
 
 
-template class SLList<Pix>;
-
 // free all unmarked pointers obtained from malloc and calloc
 void mex::cleanup(Pix ptr)
 {
   mex* context = (mex*)ptr;
-  for (Pix p = context->memlist.first(); p; context->memlist.next(p))
-    ::free(context->memlist(p));
+  for (MemSet::iterator p = context->memlist.begin();
+	p != context->memlist.end(); p++)
+    ::free(*p);
   context->memlist.clear();
 }
 
-// XXX FIXME XXX --- could this be added to class SLList<T>?
-void del(SLList<Pix>& l, const Pix& v)
-{
-  if (l.front() == v)
-    l.del_front();
-  else
-    {
-      Pix before = l.first();
-      Pix p = before;
-      l.next(p);
-      while (p && l(p) != v) 
-	{
-	  before = p; l.next(p);
-	}
-      if (p) l.del_after(before);
-    }
-}
-
-
 // mark a pointer to be freed on exit
-void mex::mark(Pix p) 
-{ 
-  if (memlist.owns(p))
+void mex::mark(Pix p)
+{
+#ifdef DEBUG
+  if (memlist.find(p) != memlist.end())
     warning("%s: double registration ignored", mexFunctionName);
-  else
-    memlist.prepend(p);
+#endif
+  memlist.insert(p);
 }
 
 // unmark a pointer to be freed on exit, either because it was
 // made persistent, or because it was already freed
-void mex::unmark(Pix p) 
-{ 
-  del(memlist, p);
+void mex::unmark(Pix p)
+{
+#ifdef DEBUG
+  if (memlist.find(p) != memlist.end())
+    warning("%s: value not marked", mexFunctionName);
+#endif
+  memlist.erase(p);
 }
 
 // allocate a pointer, and mark it to be freed on exit
@@ -314,7 +242,7 @@ Pix mex::malloc(int n)
       abort();
     }
 #endif
-  
+
   mark(ptr);
   return ptr;
 }
@@ -336,20 +264,21 @@ Pix mex::realloc(Pix ptr, int n)
   abort();
 #else
   Pix v = NULL;
-  if (n == 0) 
+  if (n == 0)
     free(ptr);
-  else if (ptr == NULL) 
+  else if (ptr == NULL)
     v = malloc(n);
   else
     {
       v = ::realloc(ptr, n);
-      if (v && memlist.owns(ptr))
+      MemSet::iterator p = memlist.find(ptr);
+      if (v && p != memlist.end())
 	{
-	  del(memlist, ptr);
-	  memlist.prepend(v);
+	  memlist.erase(p);
+	  memlist.insert(v);
 	}
     }
-#endif    
+#endif
   return v;
 }
 
@@ -365,19 +294,213 @@ void mex::free(Pix ptr)
 }
 
 
+/* ============== mxArray data type ============= */
+// Class mxArray is not much more than a struct for keeping together
+// dimensions and data.  It doesn't even ensure consistency between
+// the dimensions and the data.  Unfortunately you can't do better
+// than this without restricting the operations available in Matlab
+// for directly manipulating its mxArray type.
+
+typedef unsigned short mxChar;
+const int mxMAXNAM=64;
+
+class mxArray {
+public:
+  mxArray() {
+    nr = nc = -1;
+    pr = pi = NULL;
+    keys = NULL;
+    pmap = NULL;
+    isstr = false;
+    aname[0] = '\0';
+  }
+  ~mxArray () { 
+    if (pmap) {
+      // XXX FIXME XXX why don't string_vectors work?
+      for (int i=0; i<pmap->length(); i++) delete [] keys[i];
+      delete [] keys;
+    }
+  }
+
+  octave_value as_octave_value() const;
+
+  int rows() const { return nr; }
+  int columns() const { return nc; }
+  void rows(int r) { nr = r; }
+  void columns(int c) { nc = c; }
+  int dims() const { return 2; }
+
+  double *imag() const { return pi; }
+  double *real() const { return pr; }
+  void imag(double *p) { pi = p; }
+  void real(double *p) { pr = p; }
+
+  bool is_empty() const { return nr==0 || nc==0; }
+  bool is_numeric() const { return !isstr && (pr != NULL || nr==0 || nc==0); }
+  bool is_complex() const { return pi != NULL; }
+  bool is_sparse() const { return false; }
+  bool is_struct() const { return pmap != NULL; }
+
+  bool is_string() const { return isstr; }
+  void is_string(bool set) { isstr = set; }
+
+  const char* name() const { return aname; }
+  void name(const char *nm) {
+    strncpy(aname,nm,mxMAXNAM);
+    aname[mxMAXNAM]='\0';
+  }
+
+  // Structure support functions.
+  /* Matlab uses a fixed field order (the order in which the fields
+     were added?), but Octave uses an unordered hash for structs.  We
+     can emulate a fixed field order using pmap->keys(), which returns
+     a string_vector of key names, but these keys will not be in the
+     same order as the keys given in mxCreateStruct*.  Within the
+     creating function, we can populate the key name vector in the
+     order given, so the only problem will be those functions which
+     assume the key order is maintained between calls from Matlab.
+     Unfortunately, these might exist and I can't detect them :-(
+  */
+  // Return the map value
+  Octave_map *map(void) const { return pmap; }
+  // New structure with the given presumed field order (CreateStruct call)
+  void map(Octave_map *p, const string_vector& mapkeys) {
+    pmap = p;
+    keys = mapkeys.c_str_vec();
+  }
+  // New structure with unknown field order (passed in from Octave)
+  void map(Octave_map *p) { 
+    pmap = p;
+    if (p) keys = p->keys().c_str_vec();
+  }
+  // Get field given field name
+  mxArray* field(const std::string& key, const int index) const {
+    if (pmap && pmap->contains(key))
+#if defined(_Pix_h)
+      // 2.1.40
+      return __mex->make_value(pmap->contents(pmap->seek(key)));
+#elif defined(HAVE_OCTAVE_MAP_INDEX)
+      // 2.1.50
+      return __mex->make_value((*pmap)[key](index));
+#else
+      // 2.1.53
+      return __mex->make_value(pmap->contents(key)(index));
+#endif
+    else
+      return NULL;
+  }
+  // Set field given field name
+  void field(const std::string& key, const int index, mxArray* value) {
+#if defined(HAVE_OCTAVE_MAP_INDEX) || defined(_Pix_h)
+    // 2.1.50 || 2.1.40
+    if (pmap) (*pmap)[key](index) = value->as_octave_value();
+#else
+    // 2.1.53
+    if (pmap) 
+      pmap->assign(octave_value(index), 
+		      key, Cell(value->as_octave_value()));
+#endif
+    if (error_state) __mex->abort();
+  }
+  // Return number of fields in structure
+  int num_keys(void) const { return pmap ? pmap->length() : 0; } 
+  // Return field name from field number
+  const std::string key(const int key_num) const {
+    if (key_num >= 0 && key_num < pmap->length())
+      return keys[key_num];
+    else
+      return NULL;
+  }
+  // Return field number from field name
+  int key(const std::string &key_name) const {
+    for (int i=0; i<pmap->length(); i++) {
+      if (key_name == std::string(keys[i])) return i;
+    }
+    return -1;
+  }
+  // Get field using field number
+  mxArray* field(const int key_num, const int index) const {
+    if (key_num >= 0 && key_num < pmap->length())
+      return field(keys[key_num], index);
+    else
+      return NULL;
+  }
+  // Set field using field number
+  void field(const int key_num, const int index , mxArray* value) {
+    if (key_num >= 0 && key_num < pmap->length())
+      field(keys[key_num], index, value);
+  }
+
+private:
+  int nr, nc;
+  double *pr, *pi;
+  /* XXX FIXME XXX need to have a typeid here instead of complex logic on
+   * isstr, pmap, pr, pi, etc. */
+  Octave_map *pmap;
+  // string_vector keys;
+  char **keys;
+  bool isstr;
+  char aname[mxMAXNAM+1];
+} ;
+
+octave_value mxArray::as_octave_value() const
+{
+  octave_value ret;
+  if (isstr)
+    {
+      charMatrix chm(nr,nc);
+      char *pchm = chm.fortran_vec();
+      for (int i=0; i < nr*nc; i++)
+	pchm[i] = NINT(pr[i]);
+      ret = octave_value(chm, true);
+    }
+  else if (pmap)
+    {
+      ret = octave_value(*pmap);
+    }
+  else if (pi)
+    {
+      ComplexMatrix cm(nr, nc);
+      Complex *pcm = cm.fortran_vec();
+      for (int i=0; i < nr*nc; i++) pcm[i] = Complex(pr[i], pi[i]);
+      ret = cm;
+    }
+  else if (pr)
+    {
+      Matrix m(nr,nc);
+      double *pm = m.fortran_vec();
+      memcpy(pm, pr, nr*nc*sizeof(double));
+      ret = m;
+    }
+  else
+    ret = Matrix(0,0);
+
+  return ret;
+}
+
+
+// ====================== mex/mxArray interface ==================
+
 // make a new array value and initialize from an octave value; it will be
 // freed on exit unless marked as persistent
 mxArray* mex::make_value(const octave_value &ov)
 {
   int nr=-1, nc=-1;
   double *pr = NULL, *pi = NULL;
+  Octave_map *pmap = NULL;
 
   if (ov.is_numeric_type() || ov.is_string())
     {
       nr = ov.rows();
       nc = ov.columns();
     }
-  if (nr > 0 && nc > 0)
+  if (ov.is_map())
+    {
+      pmap = new Octave_map(ov.map_value());
+      nr = ov.rows();
+      nc = ov.columns();
+    }
+  else if (nr > 0 && nc > 0)
     {
       if (ov.is_string())
 	{
@@ -395,7 +518,7 @@ mxArray* mex::make_value(const octave_value &ov)
 	  const Complex * pz = cm.data();
 	  pr = (double *)malloc(nr*nc*sizeof(double));
 	  pi = (double *)malloc(nr*nc*sizeof(double));
-	  for (int i=0; i < nr*nc; i++) 
+	  for (int i=0; i < nr*nc; i++)
 	    {
 	      pr[i] = real(pz[i]);
 	      pi[i] = imag(pz[i]);
@@ -408,11 +531,12 @@ mxArray* mex::make_value(const octave_value &ov)
 	  memcpy(pr, m.data(), nr*nc*sizeof(double));
 	}
     }
-  
+
   mxArray *value = (mxArray*)malloc(sizeof(mxArray));
   value->is_string(ov.is_string());
   value->real(pr);
   value->imag(pi);
+  value->map(pmap);
   value->rows(nr);
   value->columns(nc);
   value->name("");
@@ -438,6 +562,34 @@ mxArray *mex::make_value(int nr, int nc, int cmplx)
   return value;
 }
 
+// make a new structure value and initialize with empty matrices
+// XXX FIXME XXX does this leak memory?  Is it persistent?
+mxArray *mex::make_value(int nr, int nc, const string_vector& keys)
+{
+  if (keys.length() == 0) return NULL;
+
+#if defined(HAVE_OCTAVE_MAP_INDEX) || defined(_Pix_h)
+  // 2.1.50 || 2.1.40
+  octave_value_list empty(nr*nc,octave_value());
+  Octave_map *pmap = new Octave_map(keys[0],empty);
+  for (int i=1; i < keys.length(); i++) 
+    pmap->assign(keys[i],empty);
+#else
+  // 2.1.53
+  Cell empty(nr,nc);
+  Octave_map *pmap = new Octave_map(keys[0],empty);
+  for (int i=1; i < keys.length(); i++)
+    pmap->assign(keys[i],empty);
+#endif
+
+  mxArray *value = (mxArray*)malloc(sizeof(mxArray));
+  value->rows(nr);
+  value->columns(nc);
+  value->map(pmap,keys);
+
+  return value;
+}
+
 // free an array and its contents
 void mex::free_value(mxArray* ptr)
 {
@@ -448,7 +600,7 @@ void mex::free_value(mxArray* ptr)
 
 // mark an array and its contents so it will not be freed on exit
 void mex::persistent(mxArray* ptr)
-{ 
+{
   persistent(Pix(ptr->real()));
   persistent(Pix(ptr->imag()));
   persistent(Pix(ptr));
@@ -457,11 +609,9 @@ void mex::persistent(mxArray* ptr)
 
 /* ========== Octave interface to mex files ============ */
 
-mex* __mex = NULL;
-
 extern "C" {
-  void F77_FCN(mexfunction,MEXFUNCTION)
-    (const int& nargout, mxArray *plhs[], 
+  void F77_FUNC(mexfunction,MEXFUNCTION)
+    (const int& nargout, mxArray *plhs[],
      const int& nargin,  mxArray *prhs[]);
   void mexFunction(const int nargout, mxArray *plhs[],
 		   const int nargin,  mxArray *prhs[]);
@@ -480,12 +630,12 @@ static void mex_exit()
 
 enum callstyle { use_fortran, use_C };
 
-octave_value_list 
+octave_value_list
 call_mex(callstyle cs, const octave_value_list& args, const int nargout)
 {
 #if 0 /* Don't bother trapping stop/exit */
   // XXX FIXME XXX ---- should really push "mex_exit" onto the octave
-  // atexit stack before we start and pop it when we are through, but 
+  // atexit stack before we start and pop it when we are through, but
   // the stack handle isn't exported from toplev.cc, so we can't.  mex_exit
   // would have to be declared as DEFUN(mex_exit,,,"") of course.
   static bool unregistered = true;
@@ -499,13 +649,14 @@ call_mex(callstyle cs, const octave_value_list& args, const int nargout)
   // nargout+1 since even for zero specified args, still want to be able
   // to return an ans.
   const int nargin = args.length();
-  mxArray * argin[nargin], * argout[nargout+1];
+  OCTAVE_LOCAL_BUFFER(mxArray*, argin, nargin);
+  OCTAVE_LOCAL_BUFFER(mxArray*, argout, nargout+1);
   for (int i=0; i < nargin; i++) argin[i] = NULL;
   for (int i=0; i < nargout+1; i++) argout[i] = NULL;
-  
+
   mex context;
   unwind_protect::add(mex::cleanup, Pix(&context));
-		      
+
   for (int i=0; i < nargin; i++) argin[i] = context.make_value(args(i));
 
   unwind_protect_ptr(__mex); // save old mex pointer
@@ -513,7 +664,7 @@ call_mex(callstyle cs, const octave_value_list& args, const int nargout)
     {
       __mex = &context;
       if (cs == use_fortran)
-	F77_FCN(mexfunction,MEXFUNCTION)(nargout, argout, nargin, argin);
+	F77_FUNC(mexfunction,MEXFUNCTION)(nargout, argout, nargin, argin);
       else
 	mexFunction(nargout, argout, nargin, argin);
     }
@@ -532,13 +683,13 @@ call_mex(callstyle cs, const octave_value_list& args, const int nargout)
   return retval;
 }
 
-octave_value_list 
+octave_value_list
 Fortran_mex(const octave_value_list& args, const int nargout)
 {
   return call_mex(use_fortran, args, nargout);
 }
 
-octave_value_list 
+octave_value_list
 C_mex(const octave_value_list& args, const int nargout)
 {
   return call_mex(use_C, args, nargout);
@@ -567,12 +718,12 @@ extern "C" {
     }
 
   // floating point representation
-  int mxIsNaN(const double v) { return xisnan(v) != 0; }
-  int mxIsFinite(const double v) { return xfinite(v) != 0; }
-  int mxIsInf(const double v) { return xisinf(v) != 0; }
+  int mxIsNaN(const double v) { return lo_ieee_is_NaN_or_NA(v) != 0; }
+  int mxIsFinite(const double v) { return lo_ieee_finite(v) != 0; }
+  int mxIsInf(const double v) { return lo_ieee_isinf(v) != 0; }
   double mxGetEps() { return DBL_EPSILON; }
-  double mxGetInf() { return octave_Inf; }
-  double mxGetNaN() { return octave_NaN; }
+  double mxGetInf() { return lo_ieee_inf_value(); }
+  double mxGetNaN() { return lo_ieee_nan_value(); }
 
   int mexEvalString(const char* s)
     {
@@ -587,7 +738,7 @@ extern "C" {
       else
 	return 0;
     }
-  int mexCallMATLAB(const int nargout, mxArray* argout[], 
+  int mexCallMATLAB(const int nargout, mxArray* argout[],
 		    const int nargin, const mxArray* argin[],
 		    const char* fname)
     {
@@ -666,15 +817,17 @@ extern "C" {
   Pix mxGetPi (const mxArray* ptr) { return ptr->imag(); }
   int mxGetM (const mxArray* ptr) { return ptr->rows(); }
   int mxGetN (const mxArray* ptr) { return ptr->columns(); }
-  void mxSetM (mxArray* ptr, const int M) { return ptr->rows(M); }
-  void mxSetN (mxArray* ptr, const int N) { return ptr->columns(N); }
+  int mxGetNumberOfDimensions (const mxArray* ptr) { return ptr->dims(); }
+  int mxGetNumberOfElements (const mxArray* ptr) { return ptr->rows()*ptr->columns(); }
+  void mxSetM (mxArray* ptr, const int M) { ptr->rows(M); }
+  void mxSetN (mxArray* ptr, const int N) { ptr->columns(N); }
   void mxSetPr (mxArray* ptr, Pix pr) { ptr->real((double *)pr); }
   void mxSetPi (mxArray* ptr, Pix pi) { ptr->imag((double *)pi); }
   double mxGetScalar (const mxArray* ptr)
     {
       double *pr =  ptr->real();
-      if (pr) return pr[0];
-      else mexErrMsgTxt("calling mxGetScalar on an empty matrix");
+      if (pr == NULL) mexErrMsgTxt("calling mxGetScalar on an empty matrix");
+      return pr[0];
     }
 
   int mxGetString (const mxArray* ptr, char *buf, int buflen)
@@ -758,7 +911,7 @@ extern "C" {
 	mexErrMsgTxt("mexPutArray: 'base' symbol table not implemented");
       else
 	mexErrMsgTxt("mexPutArray: symbol table does not exist");
-	
+      return 0;
     }
 
   mxArray *mexGetArray(const char *name, const char *space)
@@ -766,7 +919,7 @@ extern "C" {
       // XXX FIXME XXX --- this should be in variable.cc, but the correct
       // functionality is not exported.  Particularly, get_global_value()
       // generates an error if the symbol is undefined.
-      symbol_record *sr;
+      symbol_record *sr = NULL;
       if (strcmp(space,"global") == 0)
 	sr = global_sym_tab->lookup (name);
       else if (strcmp(space,"caller") == 0)
@@ -811,6 +964,41 @@ extern "C" {
       ptr->name(nm);
     }
 
+  mxArray *mxCreateStructMatrix(int nr, int nc, int num_keys, 
+				const char **keys)
+  {
+    mexWarnMsgTxt("octave mex can't yet create structures\n");
+    const string_vector ordered_keys(keys,num_keys);
+    mxArray *m = __mex->make_value(nr, nc, ordered_keys);
+    return m;
+  }
+  mxArray *mxGetField(const mxArray *ptr, int index, const char *key)
+  {
+    return ptr->field(key, index);
+  }
+  void mxSetField(mxArray *ptr, int index, const char *key, mxArray *val)
+  {
+    ptr->field(key,index,val);
+  }
+  int mxGetNumberOfFields(const mxArray* ptr) { return ptr->num_keys(); }
+  int mxIsStruct(const mxArray* ptr) { return ptr->is_struct(); }
+  const char* mxGetFieldNameByNumber(const mxArray* ptr, int key_num) 
+  {
+    return ptr->key(key_num).c_str();
+  }
+  int mxGetFieldNumber(const mxArray* ptr, const char *key)
+  {
+    return ptr->key(key);
+  }
+  mxArray* mxGetFieldByNumber(const mxArray* ptr, int index, int key_num)
+  {
+    return ptr->field(key_num,index);
+  }
+  void mxSetFieldByNumber(mxArray* ptr, int index, int key_num, mxArray* val)
+  {
+    return ptr->field(key_num,index,val);
+  }
+  
 } ;
 
 /* ============ Fortran interface to mex functions ============== */
@@ -818,144 +1006,144 @@ extern "C" {
 // fixed.  It costs and extra function call, but is easier to maintain.
 extern "C" {
 
-  void F77_FCN(mexerrmsgtxt, MEXERRMSGTXT)
+  void F77_FUNC(mexerrmsgtxt, MEXERRMSGTXT)
     (const char *s, const int slen)
     {
-      if (slen > 1 || (slen == 1 && s[0] != ' ') ) 
+      if (slen > 1 || (slen == 1 && s[0] != ' ') )
 	error("%s: %.*s", mexFunctionName, slen, s);
       else error(""); // just set the error state; don't print msg
       __mex->abort();
     }
 
-  void F77_FCN(mexprintf,MEXPRINTF)
+  void F77_FUNC(mexprintf,MEXPRINTF)
     (const char *s, const int slen)
     {
       mexPrintf("%.*s\n", slen, s);
     }
 
-  double F77_FCN(mexgeteps,MEXGETEPS)() { return mxGetEps(); }
-  double F77_FCN(mexgetinf,MEXGETINF)() { return mxGetInf(); }
-  double F77_FCN(mexgetnan,MEXGETNAN)() { return mxGetNaN(); }
-  int F77_FCN(mexisfinite,MEXISFINITE)(double v) { return mxIsFinite(v); }
-  int F77_FCN(mexisinf,MEXISINF)(double v) { return mxIsInf(v); }
-  int F77_FCN(mexisnan,MEXISNAN)(double v) { return mxIsNaN(v); }
+  double F77_FUNC(mexgeteps,MEXGETEPS)() { return mxGetEps(); }
+  double F77_FUNC(mexgetinf,MEXGETINF)() { return mxGetInf(); }
+  double F77_FUNC(mexgetnan,MEXGETNAN)() { return mxGetNaN(); }
+  int F77_FUNC(mexisfinite,MEXISFINITE)(double v) { return mxIsFinite(v); }
+  int F77_FUNC(mexisinf,MEXISINF)(double v) { return mxIsInf(v); }
+  int F77_FUNC(mexisnan,MEXISNAN)(double v) { return mxIsNaN(v); }
 
   // ====> Array access
-  Pix F77_FCN(mxcreatefull,MXCREATEFULL)
+  Pix F77_FUNC(mxcreatefull,MXCREATEFULL)
     (const int& nr, const int& nc, const int& iscomplex)
     {
       return mxCreateDoubleMatrix(nr,nc,iscomplex);
     }
 
-  void F77_FCN(mxfreematrix,MXFREEMATRIX)
+  void F77_FUNC(mxfreematrix,MXFREEMATRIX)
     (mxArray* &ptr)
     {
       mxDestroyArray(ptr);
     }
 
-  Pix F77_FCN(mxcalloc,MXCALLOC)(const int& n, const int& size)
+  Pix F77_FUNC(mxcalloc,MXCALLOC)(const int& n, const int& size)
     {
       return mxCalloc(n,size);
     }
 
-  void F77_FCN(mxfree,MXFREE)
+  void F77_FUNC(mxfree,MXFREE)
     (const Pix &ptr)
     {
       mxFree(ptr);
     }
-  
-  int F77_FCN(mxgetm,MXGETM)
-    (const mxArray* &ptr) 
-    { 
-      return mxGetM(ptr); 
+
+  int F77_FUNC(mxgetm,MXGETM)
+    (const mxArray* &ptr)
+    {
+      return mxGetM(ptr);
     }
 
-  int F77_FCN(mxgetn,MXGETN)
-    (const mxArray* &ptr) 
-    { 
-      return mxGetN(ptr); 
+  int F77_FUNC(mxgetn,MXGETN)
+    (const mxArray* &ptr)
+    {
+      return mxGetN(ptr);
     }
 
-  Pix F77_FCN(mxgetpi,MXGETPI)
-    (const mxArray* &ptr) 
+  Pix F77_FUNC(mxgetpi,MXGETPI)
+    (const mxArray* &ptr)
     {
       return mxGetPi(ptr);
     }
 
-  Pix F77_FCN(mxgetpr,MXGETPR)
-    (const mxArray* &ptr) 
+  Pix F77_FUNC(mxgetpr,MXGETPR)
+    (const mxArray* &ptr)
     {
       return mxGetPr(ptr);
     }
 
-  void F77_FCN(mxsetm,MXSETM)
-    (mxArray* &ptr, const int& m) 
-    { 
-      mxSetM(ptr, m); 
+  void F77_FUNC(mxsetm,MXSETM)
+    (mxArray* &ptr, const int& m)
+    {
+      mxSetM(ptr, m);
     }
 
-  void F77_FCN(mxsetn,MXSETN)
-    (mxArray* &ptr, const int& n) 
-    { 
+  void F77_FUNC(mxsetn,MXSETN)
+    (mxArray* &ptr, const int& n)
+    {
       mxSetN(ptr, n);
     }
 
-  void F77_FCN(mxsetpi,MXSETPI)
-    (mxArray* &ptr, Pix &pi) 
+  void F77_FUNC(mxsetpi,MXSETPI)
+    (mxArray* &ptr, Pix &pi)
     {
       mxSetPi(ptr, pi);
     }
 
-  void F77_FCN(mxsetpr,MXSETPR)
-    (mxArray* &ptr, Pix &pr) 
+  void F77_FUNC(mxsetpr,MXSETPR)
+    (mxArray* &ptr, Pix &pr)
     {
       mxSetPr(ptr, pr);
     }
-  
-  int F77_FCN(mxiscomplex,MXISCOMPLEX)
+
+  int F77_FUNC(mxiscomplex,MXISCOMPLEX)
     (const mxArray* &ptr)
     {
       return mxIsComplex(ptr);
     }
 
-  int F77_FCN(mxisdouble,MXISDOUBLE)
+  int F77_FUNC(mxisdouble,MXISDOUBLE)
     (const mxArray* &ptr)
     {
       return mxIsDouble(ptr);
     }
-  
-  int F77_FCN(mxisnumeric,MXISNUMERIC)
+
+  int F77_FUNC(mxisnumeric,MXISNUMERIC)
     (const mxArray* &ptr)
     {
       return mxIsNumeric(ptr);
     }
-  
-  int F77_FCN(mxisfull,MXISFULL)
+
+  int F77_FUNC(mxisfull,MXISFULL)
     (const mxArray* &ptr)
     {
       return 1 - mxIsSparse(ptr);
     }
-  
-  int F77_FCN(mxissparse,MXISSPARSE)
+
+  int F77_FUNC(mxissparse,MXISSPARSE)
     (const mxArray* &ptr)
     {
       return mxIsSparse(ptr);
     }
-  
-  int F77_FCN(mxisstring,MXISSTRING)
+
+  int F77_FUNC(mxisstring,MXISSTRING)
     (const mxArray* &ptr)
     {
       return mxIsChar(ptr);
     }
 
-  int F77_FCN(mxgetstring,MXGETSTRING)
+  int F77_FUNC(mxgetstring,MXGETSTRING)
     (const mxArray* &ptr, char *str, const int& len)
     {
       return mxGetString(ptr, str, len);
     }
 
-  int F77_FCN(mexcallmatlab,MEXCALLMATLAB)
-    (const int& nargout, mxArray** argout, 
+  int F77_FUNC(mexcallmatlab,MEXCALLMATLAB)
+    (const int& nargout, mxArray** argout,
      const int& nargin, const mxArray** argin,
      const char* fname,
      const int fnamelen)
@@ -967,23 +1155,23 @@ extern "C" {
     }
 
   // ======> Fake pointer support
-  void F77_FCN(mxcopyreal8toptr,MXCOPYREAL8TOPTR)
+  void F77_FUNC(mxcopyreal8toptr,MXCOPYREAL8TOPTR)
     (const double *d, const int& prref, const int& len)
     {
       TRACEFN;
       double *pr = (double *)prref;
       for (int i=0; i < len; i++) pr[i] = d[i];
     }
-  
-  void F77_FCN(mxcopyptrtoreal8,MXCOPYPTRTOREAL8)
+
+  void F77_FUNC(mxcopyptrtoreal8,MXCOPYPTRTOREAL8)
     (const int& prref, double *d, const int& len)
     {
       TRACEFN;
       double *pr = (double *)prref;
       for (int i=0; i < len; i++) d[i] = pr[i];
     }
-  
-  void F77_FCN(mxcopycomplex16toptr,MXCOPYCOMPLEX16TOPTR)
+
+  void F77_FUNC(mxcopycomplex16toptr,MXCOPYCOMPLEX16TOPTR)
     (const double *d, int& prref, int& piref, const int& len)
     {
       TRACEFN;
@@ -991,8 +1179,8 @@ extern "C" {
       double *pi = (double *)piref;
       for (int i=0; i < len; i++) pr[i] = d[2*i], pi[i] = d[2*i+1];
     }
-  
-  void F77_FCN(mxcopyptrtocomplex16,MXCOPYPTRTOCOMPLEX16)
+
+  void F77_FUNC(mxcopyptrtocomplex16,MXCOPYPTRTOCOMPLEX16)
     (const int& prref, const int& piref, double *d, const int& len)
     {
       TRACEFN;
@@ -1000,5 +1188,5 @@ extern "C" {
       double *pi = (double *)piref;
       for (int i=0; i < len; i++) d[2*i]=pr[i], d[2*i+1] = pi[i];
     }
-  
+
 } ;
