@@ -32,7 +32,15 @@
  *************************************************************************/
 
 #include <octave/oct.h>
-#include <math.h>
+#include <cmath>
+
+#ifndef OCTAVE_LOCAL_BUFFER
+#include <vector>
+#define OCTAVE_LOCAL_BUFFER(T, buf, size) \
+  std::vector<T> buf ## _vector (size); \
+  T *buf = &(buf ## _vector[0])
+#endif
+
 
 #define CONST const
 #define BANDPASS       1
@@ -750,7 +758,7 @@ int remez(double h[], int numtaps,
 
 DEFUN_DLD (remez, args, ,
   "b = remez(n, f, a [, w] [, ftype] [, griddensity])\n\
-\n\
+Parks-McClellan optimal FIR filter design.\n\
 n gives the number of taps in the returned filter\n\
 f gives frequency at the band edges [ b1 e1 b2 e2 b3 e3 ...]\n\
 a gives amplitude at the band edges [ a(b1) a(e1) a(b2) a(e2) ...]\n\
@@ -779,7 +787,7 @@ Frequency is in the range (0, 1), with 1 being the nyquist frequency")
 
   ColumnVector o_bands(args(1).vector_value());
   int numbands = o_bands.length()/2;
-  double bands[numbands*2];
+  OCTAVE_LOCAL_BUFFER(double, bands, numbands*2);
   if (numbands < 1 || o_bands.length()%2 == 1) {
     error("remez: must have an even number of band edges");
     return retval;
@@ -797,16 +805,16 @@ Frequency is in the range (0, 1), with 1 being the nyquist frequency")
   for(i=0; i < 2*numbands; i++) bands[i] = o_bands(i)/2.0;
 
   ColumnVector o_response(args(2).vector_value());
-  double response[numbands*2];
+  OCTAVE_LOCAL_BUFFER (double, response, numbands*2);
   if (o_response.length() != o_bands.length()) {
     error("remez: must have one response magnitude for each band edge");
     return retval;
   }
   for(i=0; i < 2*numbands; i++) response[i] = o_response(i);
 
-  string stype = string("bandpass");
+  std::string stype = std::string("bandpass");
   int density = 16;
-  double weight[numbands];
+  OCTAVE_LOCAL_BUFFER (double, weight, numbands);
   for (i=0; i < numbands; i++) weight[i] = 1.0;
   if (nargin > 3) {
     if (args(3).is_real_matrix()) {
@@ -864,7 +872,7 @@ Frequency is in the range (0, 1), with 1 being the nyquist frequency")
     return retval;
   }
 
-  double coeff[numtaps+5];
+  OCTAVE_LOCAL_BUFFER (double, coeff, numtaps+5);
   int err = remez(coeff,numtaps,numbands,bands,response,weight,itype,density);
 
   if (err == -1)
@@ -883,3 +891,26 @@ Frequency is in the range (0, 1), with 1 being the nyquist frequency")
 
   return octave_value(h);
 }
+
+/*
+%!test
+%! b = [
+%!    0.0415131831103279
+%!    0.0581639884202646
+%!   -0.0281579212691008
+%!   -0.0535575358002337
+%!   -0.0617245915143180
+%!    0.0507753178978075
+%!    0.2079018331396460
+%!    0.3327160895375440
+%!    0.3327160895375440
+%!    0.2079018331396460
+%!    0.0507753178978075
+%!   -0.0617245915143180
+%!   -0.0535575358002337
+%!   -0.0281579212691008
+%!    0.0581639884202646
+%!    0.0415131831103279];
+%! assert(remez(15,[0,0.3,0.4,1],[1,1,0,0]),b,1e-14);
+
+ */
