@@ -24,7 +24,7 @@ this program with any Open Source program, as defined by the
 Open Source Initiative (www.opensource.org)
 
 
-$Id: complex_sparse_ops.cc,v 1.3 2001/10/14 03:06:31 aadler Exp $
+$Id: complex_sparse_ops.cc,v 1.23 2003/12/22 15:13:23 pkienzle Exp $
 
 */
 
@@ -68,18 +68,28 @@ create_SuperMatrix( int nr, int nc, int nnz,
    return X;
 }
 
+// this is required so that code that can't include
+// complex sparse headers (because they conflict)
+// can still access this
+Complex * new_SuperLU_Complex( int size ) {
+    return (Complex *) doublecomplexMalloc( size );
+}
+
 //
 // Octave complex sparse methods
 //
 
-inline
 octave_complex_sparse::octave_complex_sparse (SuperMatrix A )
 {
    DEBUGMSG("complex_sparse( SuperMatrix A)");
    X= A;
 }
 
-inline
+octave_complex_sparse::octave_complex_sparse (void )
+{
+   DEBUGMSG("complex_sparse( void)");
+}
+
 octave_complex_sparse::~octave_complex_sparse (void)
 {
    DEBUGMSG("complex_sparse destructor");
@@ -88,28 +98,27 @@ octave_complex_sparse::~octave_complex_sparse (void)
 
 //NOTE: I'm not sure when this will get called,
 //      so I don't know what to do
-inline
 octave_complex_sparse::octave_complex_sparse (const octave_complex_sparse& S)
 {
    DEBUGMSG("complex_sparse copy-constructor");
    X= S.super_matrix();
 }   
 
-inline octave_value *
-octave_complex_sparse::clone (void)
+octave_value *
+octave_complex_sparse::clone (void) const
 {
    DEBUGMSG("complex_sparse - clone");
    return new octave_complex_sparse (*this);
 }
 
-inline octave_complex_sparse
-octave_complex_sparse::sparse_value (bool = false) const {
+octave_complex_sparse
+octave_complex_sparse::sparse_value (bool) const {
    DEBUGMSG("complex_sparse_value");
    return  (*this);
 }
 
 SuperMatrix
-octave_complex_sparse::super_matrix (bool = false) const {
+octave_complex_sparse::super_matrix (bool) const {
    return X;
 }
 
@@ -120,6 +129,9 @@ int octave_complex_sparse::rows    (void) const {
 int octave_complex_sparse::columns (void) const {
    return X.ncol;
 }
+int octave_complex_sparse::cols (void) const {
+   return X.ncol;
+}
 
 int octave_complex_sparse::nnz     (void) const {
    NCformat * NCF  = (NCformat * ) X.Store;
@@ -128,7 +140,7 @@ int octave_complex_sparse::nnz     (void) const {
 
 // upconvert octave_sparse to octave_complex_sparse
 octave_complex_sparse
-octave_sparse::complex_sparse_value (bool = false) const {
+octave_sparse::complex_sparse_value (bool) const {
    DEBUGMSG("sparse - complex_sparse_value");
    DEFINE_SP_POINTERS_REAL( X )
    int nnz= NCFX->nnz;
@@ -156,7 +168,8 @@ oct_complex_sparse_to_full ( SuperMatrix X ) {
    
    ComplexMatrix M( Xnr, Xnc );
    for (int j=0; j< Xnc; j++) {
-   // I think new Matrices are initialized to zero, however just in case
+      OCTAVE_QUIT;
+      // I think new Matrices are initialized to zero, however just in case
       for (int i=0; i< Xnr; i++) M(i,j)= 0;
    
       for (int i= cidxX[j]; i< cidxX[j+1]; i++) 
@@ -184,16 +197,16 @@ octave_complex_sparse::numeric_conversion_function (void) const
 
 //idx_vector index_vector (void) const { return idx_vector ((double) iv); }
 
-octave_value octave_complex_sparse::any (void) const {
+octave_value octave_complex_sparse::any (int dim) const {
    DEBUGMSG("complex_sparse - any");
    ComplexMatrix M= oct_complex_sparse_to_full( X );
-   return M.any();
+   return M.any(dim);
 }
 
-octave_value octave_complex_sparse::all (void) const {
+octave_value octave_complex_sparse::all (int dim) const {
    DEBUGMSG("complex_sparse - all");
    ComplexMatrix M= oct_complex_sparse_to_full( X );
-   return M.all();
+   return M.all(dim);
 }
 
 bool octave_complex_sparse::is_defined    (void) const  { return true; }
@@ -218,7 +231,7 @@ bool octave_complex_sparse::is_true (void) const {
 // rebuild a full matrix from a sparse one
 // this functionality is accessed through 'full'
 ComplexMatrix
-octave_complex_sparse::complex_matrix_value (bool = false) const {
+octave_complex_sparse::complex_matrix_value (bool) const {
    DEBUGMSG("complex_sparse - complex_matrix_value");
    ComplexMatrix M= oct_complex_sparse_to_full( X );
    return M;
@@ -260,9 +273,11 @@ oct_complex_sparse_transpose ( SuperMatrix X ) {
 
    DECLARE_SP_POINTERS_CPLX( B )
 
+   BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    zCompRow_to_CompCol( Xnc, Xnr, nnz,
                   (doublecomplex *) coefX, ridxX, cidxX,
                   (doublecomplex **) &coefB, &ridxB, &cidxB);
+   END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    
    SuperMatrix B= create_SuperMatrix( Xnc, Xnr, nnz, coefB, ridxB, cidxB );
    return B;
@@ -302,7 +317,6 @@ UNOPDECL (hermitian, a)
    return v.hermitian();
 } // hermitian
 
-#if 0
 typedef struct { long val;
                  long idx; } sort_idx;   
 // comparison function for sort in index
@@ -317,6 +331,7 @@ ixp_comp(const void *i,const void*j )
 static inline void
 sort_with_idx (sort_idx * sidx, const idx_vector& idxv, long ixl) 
 {
+   BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    if (idxv.is_colon() ) {
       for (int i=0; i< ixl; i++) {
          sidx[i].val= i;
@@ -331,11 +346,12 @@ sort_with_idx (sort_idx * sidx, const idx_vector& idxv, long ixl)
 
       qsort( sidx, ixl, sizeof(sort_idx), ixp_comp );
    }
+   END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
 }   
 
 // Return a full vector output
 // Does it make sense to output a sparse matrix here?
-static ColumnVector
+static ComplexColumnVector
 sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
    DEBUGMSG("complex_sparse_index_oneidx");
    DEFINE_SP_POINTERS_CPLX( X )
@@ -346,18 +362,21 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
    else  
       ixl= ix.length(-1); 
 
-   sort_idx ixp[ ixl ];
+   OCTAVE_LOCAL_BUFFER (sort_idx, ixp, ixl );
    sort_with_idx (ixp, ix, ixl);
 
-   ColumnVector O( ixl );
+   ComplexColumnVector O( ixl );
    long ip= -Xnr; // previous column position
    long jj=0,jl=0;
    for (long k=0; k< ixl; k++) {
+      OCTAVE_QUIT;
       long ii  = ixp[k].val;
       long kout= ixp[k].idx;
 
-      if ( ii<0 || ii>=Xnr*Xnc) 
-         SP_FATAL_ERR("invalid matrix index");
+      if ( ii<0 || ii>=Xnr*Xnc) {
+         error("sparse index out of range");
+	 return O;
+      }
  
       int rown= ii/Xnr;
       if ( rown > ip/Xnr ) { // we've moved to a new column
@@ -365,9 +384,9 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
          jj= cidxX[rown+1];
       }
 
-      while ( ridxX[jl] < ii%Xnr && jl < jj ) jl++;
+      while ( jl < jj && ridxX[jl] < ii%Xnr ) jl++;
 
-      if ( ridxX[jl] == ii%Xnr && jl<jj ) 
+      if ( jl<jj && ridxX[jl] == ii%Xnr ) 
          O( kout ) = coefX[jl] ;
       else
          O( kout ) = 0 ;
@@ -378,7 +397,7 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
 } // sparse_index_oneidx (
 
 
-static SuperMatrix
+static octave_value
 sparse_index_twoidx ( SuperMatrix X,
                       const idx_vector ix,
                       const idx_vector jx) {
@@ -392,17 +411,17 @@ sparse_index_twoidx ( SuperMatrix X,
    if (jx.is_colon() )      jxl= Xnc;
    else                     jxl= jx.length(-1); 
 
-   sort_idx ixp[ ixl ];
+   OCTAVE_LOCAL_BUFFER (sort_idx, ixp, ixl );
    sort_with_idx (ixp, ix, ixl);
 
    // extimate the nnz in the output matrix
    int nnz = (int) ceil( (NCFX->nnz) * (1.0*ixl / Xnr) * (1.0*jxl / Xnc) ); 
 
-   double * coefB = doubleMalloc(nnz);
+   Complex * coefB = (Complex *) doublecomplexMalloc(nnz);
    int    * ridxB = intMalloc   (nnz);
    int    * cidxB = intMalloc   (jxl+1);  cidxB[0]= 0;
 
-   double tcol[ixl];  // a column of the extracted matrix
+   OCTAVE_LOCAL_BUFFER (Complex, tcol, ixl );  // a column of the extracted matrix
 
    int cx= 0, ll=0;
    int ip= -Xnc; // previous column position
@@ -410,17 +429,24 @@ sparse_index_twoidx ( SuperMatrix X,
       if (jx.is_colon() )    ll= l;
       else                   ll= jx(l);
 
-      if ( ll<0 || ll>=Xnc) 
-            SP_FATAL_ERR("invalid matrix index (x index)");
+      if ( ll<0 || ll>=Xnc) {
+	 // XXX FIXME XXX memory leak coefB,cidxB,ridxB
+	 error("sparse column index out of range");
+	 return octave_value();
+      }
 
       int jl= cidxX[ll];
       int jj= cidxX[ll+1];
       for (long k=0; k< ixl; k++) {
+	 OCTAVE_QUIT;
          long ii  = ixp[k].val;
          long kout= ixp[k].idx;
    
-         if ( ii<0 || ii>=Xnr) 
-            SP_FATAL_ERR("invalid matrix index (x index)");
+         if ( ii<0 || ii>=Xnr) {
+	   // XXX FIXME XXX memory leak coefB,cidxB,ridxB
+	   error("sparse row index out of range");
+	   return octave_value();
+	 }
 
          while ( ridxX[jl] < ii && jl < jj ) jl++;
 
@@ -434,7 +460,7 @@ sparse_index_twoidx ( SuperMatrix X,
    
       } // for k
       for (int j=0; j<ixl; j++) {
-         if (tcol[j] !=0 ) {
+         if (tcol[j] != Complex(0) ) {
             check_bounds( cx, nnz, ridxB, coefB );
             ridxB[cx]= j;
             coefB[cx]= tcol[j];
@@ -447,22 +473,52 @@ sparse_index_twoidx ( SuperMatrix X,
    maybe_shrink( cx, nnz, ridxB, coefB );
 
    SuperMatrix B;
-   dCreate_CompCol_Matrix(&B, ixl, jxl, cx,
-                       coefB, ridxB, cidxB, NC, _D, GE);
 
-   return B;                          
+   zCreate_CompCol_Matrix(&B, ixl, jxl, cx,
+         (doublecomplex *) coefB, ridxB, cidxB, NC, _D, GE);
+
+   return octave_value(new octave_complex_sparse(B));
 } // sparse_index_twoidx (
 
-// indexing operations
+
 octave_value_list
-octave_complex_sparse::do_multi_index_op (int, const octave_value_list& idx) 
+octave_complex_sparse::subsref( const std::string SUBSREF_STRREF type,
+                        const LIST<octave_value_list>& idx,
+                        int nargout)
+{
+   octave_value_list retval;
+   switch (type[0]) {
+     case '(':
+       retval = do_index_op (idx.front (), 0);
+       break;
+
+     case '{':
+     case '.':
+     {
+       std::string nm = type_name ();
+       error ("%s cannot be indexed with %c", nm.c_str (), type[0]);
+     }
+     break;
+
+     default:
+       panic_impossible ();
+   }
+
+   if (idx.LISTSIZE () > 1)
+       retval = retval(0).next_subsref (type, idx);
+
+   return retval;
+}
+
+octave_value
+octave_complex_sparse::do_index_op ( const octave_value_list& idx, int resize_ok) 
 {
    DEBUGMSG("complex_sparse - index op");
    octave_value retval;
    
    if ( idx.length () == 1) {
       const idx_vector ix = idx (0).index_vector ();
-      ColumnVector O= sparse_index_oneidx( X, ix );
+      ComplexColumnVector O= sparse_index_oneidx( X, ix );
 
       // the rules are complicated here X(Y):
       // X is matrix: result is same shape as Y
@@ -477,10 +533,9 @@ octave_complex_sparse::do_multi_index_op (int, const octave_value_list& idx)
       const idx_vector ix = idx (0).index_vector ();
       const idx_vector jx = idx (1).index_vector ();
 
-      retval= new octave_complex_sparse ( 
-                  sparse_index_twoidx ( X, ix, jx ));
+      retval= sparse_index_twoidx ( X, ix, jx );
    } else
-      SP_FATAL_ERR("need 1 or 2 indices for sparse indexing operations");
+      error("need 1 or 2 indices for sparse indexing operations");
 
    return retval;
 } // octave_complex_sparse::do_index_op
@@ -501,15 +556,16 @@ octave_complex_sparse::extract (int r1, int c1, int r2, int c2) const {
    int nnz = (int) ceil( (NCFX->nnz) * (1.0*m / Xnr)
                                      * (1.0*n / Xnc) ); 
 
-   double * coefB = doubleMalloc(nnz);
+   Complex * coefB = (Complex *) doublecomplexMalloc(nnz);
    int    * ridxB = intMalloc   (nnz);
    int    * cidxB = intMalloc   (n+1);  cidxB[0]= 0;
 
    int cx= 0;
    for (int i=0, ii= c1; i < n ; i++, ii++) {
+      OCTAVE_QUIT;
       for ( int j= cidxX[ii]; j< cidxX[ii+1]; j++) {
          int row = ridxX[ j ];
-         if ( row>= r1 && row<=r2 && coefX[j] !=0 ) {
+         if ( row>= r1 && row<=r2 && coefX[j] != Complex(0) ) {
             check_bounds( cx, nnz, ridxB, coefB );
             ridxB[ cx ]= row - r1;
             coefB[ cx ]= coefX[ j ];
@@ -523,16 +579,15 @@ octave_complex_sparse::extract (int r1, int c1, int r2, int c2) const {
    maybe_shrink( cx, nnz, ridxX, coefX );
 
    SuperMatrix B;
-   dCreate_CompCol_Matrix(&B, m, n, cx,
-                          coefB, ridxB, cidxB, NC, _D, GE);
+   zCreate_CompCol_Matrix(&B, m, n, cx,
+        (doublecomplex *) coefB, ridxB, cidxB, NC, _D, GE);
 
    return new octave_complex_sparse ( B );
 } // octave_complex_sparse::extract (int r1, int c1, int r2, int c2) const {
 
-#endif
 
 void
-octave_complex_sparse::print (ostream& os, bool pr_as_read_syntax ) const
+octave_complex_sparse::print (std::ostream& os, bool pr_as_read_syntax ) const
 {
    DEBUGMSG("complex_sparse - print");
 #if 0
@@ -548,9 +603,13 @@ octave_complex_sparse::print (ostream& os, bool pr_as_read_syntax ) const
    // add one to the printed indices to go from
    //  zero-based to one-based arrays
    for (int j=0; j< Xnc; j++) 
-      for (int i= cidxX[j]; i< cidxX[j+1]; i++) 
+     for (int i= cidxX[j]; i< cidxX[j+1]; i++) {
+         OCTAVE_QUIT;
          os << "  (" << ridxX[i]+1 <<
-               " , "  << j+1 << ") -> " << coefX[i] << "\n";
+               " , "  << j+1 << ") -> ";
+	 octave_print_internal( os, coefX[i], false );
+	 os << "\n";
+     }
 #endif                  
 } // print
 
@@ -564,10 +623,212 @@ complex_sparse_complex_multiply (
             const octave_complex&        cplx)
 {
   DEBUGMSG("complex_sparse - complex_sparse_complex_multiply");
-  Complex c= cplx.complex_value();
+  Complex s= cplx.complex_value();
 
   SuperMatrix X= spar.super_matrix();
   DEFINE_SP_POINTERS_CPLX( X )
+  int source_nnz= NCFX->nnz;
+
+  Complex *coefB = (Complex *) doublecomplexMalloc(source_nnz);
+  int *    ridxB = intMalloc(source_nnz);
+  int *    cidxB = intMalloc(X.ncol+1);
+
+  int nnz=0;
+  int col=0;
+  for ( int i=0; i< source_nnz; i++) {
+     Complex v= coefX[i] * s;
+     if (v==0.) continue;
+     coefB[nnz]= v;
+     ridxB[nnz]= ridxX[i];
+     while(i>=cidxX[col]) cidxB[col++] = nnz;
+     nnz++;
+  }
+
+  while(col <= Xnc) cidxB[col++] = nnz;
+  maybe_shrink( source_nnz, nnz, ridxX, coefX );
+  SuperMatrix B= create_SuperMatrix( Xnr, Xnc, nnz, coefB, ridxB, cidxB );
+  return new octave_complex_sparse ( B );
+}
+
+//
+// complex_sparse by scalar operations
+//
+
+DEFBINOP (cs_n_mul, complex_sparse, scalar) {
+  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_scalar&);
+  Complex cv( v2.double_value(), 0 );
+  return complex_sparse_complex_multiply (v1, cv);
+}  
+
+DEFBINOP (n_cs_mul,  scalar, complex_sparse) {
+  CAST_BINOP_ARGS (const octave_scalar&, const octave_complex_sparse&);
+  Complex cv( v1.double_value(), 0 );
+  return complex_sparse_complex_multiply (v2, cv);
+}  
+
+DEFBINOP (cs_n_div, complex_sparse, scalar) {
+  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_scalar&);
+  Complex cv ( v2.double_value (), 0);
+  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
+  return complex_sparse_complex_multiply (v1, 1.0 / cv );
+}  
+
+DEFBINOP (n_cs_ldiv, scalar, complex_sparse) {
+  CAST_BINOP_ARGS (const octave_scalar&, const octave_complex_sparse&);
+  Complex cv ( v1.double_value (), 0);
+  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
+  return complex_sparse_complex_multiply (v2, 1.0 / cv );
+}  
+
+//
+// sparse by complex operations
+//
+
+octave_value
+sparse_complex_multiply (
+            const octave_sparse&  spar,
+            const octave_complex& cplx)
+{
+  DEBUGMSG("complex_sparse - sparse_complex_multiply");
+  Complex s = cplx.complex_value();
+
+  SuperMatrix X= spar.super_matrix();
+  DEFINE_SP_POINTERS_REAL( X )
+  int source_nnz= NCFX->nnz;
+
+  Complex *coefB = (Complex *) doublecomplexMalloc(source_nnz);
+  int *    ridxB = intMalloc(source_nnz);
+  int *    cidxB = intMalloc(X.ncol+1);
+
+  int nnz=0;
+  int col=0;
+  for ( int i=0; i< source_nnz; i++) {
+     coefB[nnz]=  coefX[i] * s;
+     // Use volatile comparison to force zero
+     volatile double re = coefB[nnz].real();
+     volatile double im = coefB[nnz].imag();
+     if (re==0. && im==0.) continue;
+     ridxB[nnz] = ridxX[i];
+     while(i>=cidxX[col]) cidxB[col++] = nnz;
+     nnz++;
+  }
+
+  while(col <= Xnc) cidxB[col++] = nnz;
+  maybe_shrink( source_nnz, nnz, ridxX, coefX );
+  SuperMatrix B= create_SuperMatrix( Xnr, Xnc, nnz, coefB, ridxB, cidxB );
+  return new octave_complex_sparse ( B );
+}
+
+DEFBINOP (s_c_mul, sparse, complex) {
+  CAST_BINOP_ARGS (const octave_sparse&, const octave_complex&);
+  return sparse_complex_multiply (v1, v2);
+}  
+
+DEFBINOP (c_s_mul, complex, sparse) {
+  CAST_BINOP_ARGS (const octave_complex&, const octave_sparse&);
+  return sparse_complex_multiply (v2, v1);
+}  
+
+DEFBINOP (s_c_div, sparse, complex) {
+  CAST_BINOP_ARGS (const octave_sparse&, const octave_complex&);
+  Complex cv= v2.complex_value();
+  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
+  return sparse_complex_multiply (v1, 1.0 / cv );
+}  
+
+DEFBINOP (c_s_ldiv, complex, sparse) {
+  CAST_BINOP_ARGS (const octave_complex&, const octave_sparse&);
+  Complex cv= v1.complex_value();
+  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
+  return sparse_complex_multiply (v2, 1.0 / cv );
+}  
+
+//
+// complex_sparse by complex operations
+//
+
+DEFBINOP (cs_c_mul, complex_sparse, complex) {
+  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_complex&);
+  return complex_sparse_complex_multiply (v1, v2);
+}  
+
+DEFBINOP (c_cs_mul, complex, complex_sparse) {
+  CAST_BINOP_ARGS (const octave_complex&, const octave_complex_sparse&);
+  return complex_sparse_complex_multiply (v2, v1);
+}  
+
+DEFBINOP (cs_c_div, complex_sparse, complex) {
+  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_complex&);
+  Complex cv= v2.complex_value();
+  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
+  return complex_sparse_complex_multiply (v1, 1.0 / cv );
+}  
+
+DEFBINOP (c_cs_ldiv, complex, complex_sparse) {
+  CAST_BINOP_ARGS (const octave_complex&, const octave_complex_sparse&);
+  Complex cv= v1.complex_value();
+  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
+  return complex_sparse_complex_multiply (v2, 1.0 / cv );
+}  
+
+// complex exponentition
+
+octave_value
+complex_sparse_complex_power (
+            const octave_complex_sparse& spar,
+            const octave_complex&        cplx)
+{
+  DEBUGMSG("complex_sparse - complex_sparse_complex_power");
+  Complex s= cplx.complex_value();
+
+  SuperMatrix X= spar.super_matrix();
+  DEFINE_SP_POINTERS_CPLX( X )
+  int source_nnz= NCFX->nnz;
+
+  Complex *coefB = (Complex *) doublecomplexMalloc(source_nnz);
+  int *    ridxB = intMalloc(source_nnz);
+  int *    cidxB = intMalloc(X.ncol+1);
+
+  int nnz=0;
+  int col=0;
+  for ( int i=0; i< source_nnz; i++) {
+     // Use volatile comparison to force zero
+     coefB[nnz] = std::pow( coefX[i] , s);
+     volatile double re = coefB[nnz].real();
+     volatile double im = coefB[nnz].imag();
+     if (re==0. && im==0.) continue;
+     ridxB[nnz]= ridxX[i];
+     while(i>=cidxX[col]) cidxB[col++] = nnz;
+     nnz++;
+  }
+
+  while(col <= Xnc) cidxB[col++] = nnz;
+  maybe_shrink( source_nnz, nnz, ridxX, coefX );
+  SuperMatrix B= create_SuperMatrix( Xnr, Xnc, nnz, coefB, ridxB, cidxB );
+  return new octave_complex_sparse ( B );
+}
+
+DEFBINOP (cs_n_pow, complex_sparse, scalar) {
+  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_scalar&);
+  Complex cv( v2.double_value(), 0 );
+  return complex_sparse_complex_power (v1, cv);
+}  
+
+DEFBINOP (cs_c_pow, complex_sparse, complex) {
+  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_complex&);
+  return complex_sparse_complex_power (v1, v2);
+}  
+
+octave_value
+sparse_complex_power (
+            const octave_sparse&   spar,
+            const octave_complex&  cplx)
+{
+  DEBUGMSG("complex_sparse - sparse_complex_power");
+  Complex c= cplx.complex_value();
+
+  SuperMatrix X= spar.super_matrix();
+  DEFINE_SP_POINTERS_REAL( X )
   int nnz= NCFX->nnz;
 
   Complex *coefB = (Complex *) doublecomplexMalloc(nnz);
@@ -578,7 +839,8 @@ complex_sparse_complex_multiply (
      cidxB[i]=  cidxX[i];
 
   for ( int i=0; i< nnz; i++) {
-     coefB[i]=  coefX[i] * c;
+     Complex cv( coefX[i], 0 );
+     coefB[i]=  std::pow (cv , c);
      ridxB[i]=  ridxX[i];
   }
 
@@ -586,62 +848,9 @@ complex_sparse_complex_multiply (
   return new octave_complex_sparse ( B );
 }
 
-//
-// complex_sparse by scalar operations
-//
-
-DEFBINOP (s_n_mul, complex_sparse, scalar) {
-  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_scalar&);
-  Complex cv( v2.double_value(), 0 );
-  return complex_sparse_complex_multiply (v1, cv);
-}  
-
-DEFBINOP (n_s_mul,  scalar, complex_sparse) {
-  CAST_BINOP_ARGS (const octave_scalar&, const octave_complex_sparse&);
-  Complex cv( v1.double_value(), 0 );
-  return complex_sparse_complex_multiply (v2, cv);
-}  
-
-DEFBINOP (s_n_div, complex_sparse, scalar) {
-  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_scalar&);
-  Complex cv ( v2.double_value (), 0);
-  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
-  return complex_sparse_complex_multiply (v1, 1.0 / cv );
-}  
-
-DEFBINOP (n_s_ldiv, scalar, complex_sparse) {
-  CAST_BINOP_ARGS (const octave_scalar&, const octave_complex_sparse&);
-  Complex cv ( v1.double_value (), 0);
-  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
-  return complex_sparse_complex_multiply (v2, 1.0 / cv );
-}  
-
-//
-// complex_sparse by complex operations
-//
-
-DEFBINOP (s_c_mul, complex_sparse, complex) {
-  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_complex&);
-  return complex_sparse_complex_multiply (v1, v2);
-}  
-
-DEFBINOP (c_s_mul, complex, complex_sparse) {
-  CAST_BINOP_ARGS (const octave_complex&, const octave_complex_sparse&);
-  return complex_sparse_complex_multiply (v2, v1);
-}  
-
-DEFBINOP (s_c_div, complex_sparse, complex) {
-  CAST_BINOP_ARGS (const octave_complex_sparse&, const octave_complex&);
-  Complex cv= v2.complex_value();
-  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
-  return complex_sparse_complex_multiply (v1, 1.0 / cv );
-}  
-
-DEFBINOP (c_s_ldiv, complex, complex_sparse) {
-  CAST_BINOP_ARGS (const octave_complex&, const octave_complex_sparse&);
-  Complex cv= v1.complex_value();
-  if (cv == Complex(0,0) ) gripe_divide_by_zero ();
-  return complex_sparse_complex_multiply (v2, 1.0 / cv );
+DEFBINOP (s_c_pow, sparse, complex) {
+  CAST_BINOP_ARGS (const octave_sparse&, const octave_complex&);
+  return sparse_complex_power (v1, v2);
 }  
 
 //
@@ -976,7 +1185,7 @@ DEFBINOP( s_cs_ldiv, sparse, complex_sparse) {
    DEBUGMSG("sparse - s_cs_ldiv");
    CAST_BINOP_ARGS ( const octave_sparse&, const octave_complex_sparse&);
    int n = v1.columns();
-   int perm_c[n];
+   OCTAVE_LOCAL_BUFFER (int, perm_c, n );
    int permc_spec=3;
    octave_value_list Si= oct_sparse_inverse( v1, perm_c, permc_spec );
    octave_value inv= Si(0)*Si(1)*Si(2)*Si(3);
@@ -993,7 +1202,7 @@ DEFBINOP( cs_s_ldiv, complex_sparse, sparse) {
    CAST_BINOP_ARGS ( const octave_complex_sparse&, const octave_sparse&);
    SuperMatrix   B= v2.super_matrix(); DEFINE_SP_POINTERS_REAL( B )
    int n = v1.columns();
-   int perm_c[n];
+   OCTAVE_LOCAL_BUFFER (int, perm_c, n );
    int permc_spec=3;
    octave_value_list Si= oct_sparse_inverse( v1, perm_c, permc_spec );
    octave_value inv= Si(0)*Si(1)*Si(2)*Si(3);
@@ -1009,7 +1218,7 @@ DEFBINOP( cs_cs_ldiv, complex_sparse, complex_sparse) {
    CAST_BINOP_ARGS ( const octave_complex_sparse&, const octave_complex_sparse&);
    SuperMatrix   B= v2.super_matrix(); DEFINE_SP_POINTERS_CPLX( B )
    int n = v1.columns();
-   int perm_c[n];
+   OCTAVE_LOCAL_BUFFER (int, perm_c, n );
    int permc_spec=3;
    octave_value_list Si= oct_sparse_inverse( v1, perm_c, permc_spec );
    octave_value inv= Si(0)*Si(1)*Si(2)*Si(3);
@@ -1037,26 +1246,28 @@ do_cs_cf_ldiv( SuperMatrix A, ComplexMatrix& M )
    if (Anc != Bnr) {
       gripe_nonconformant ("operator \\", Anr, Anc, Bnr, Bnc);
    } else {
+      int permc_spec = 3;
+      OCTAVE_LOCAL_BUFFER (int, perm_c, Anc );
+      OCTAVE_LOCAL_BUFFER (int, perm_r, Anr );
+      BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
       assert (Anc == Bnr);
       SuperMatrix L,U,B;
       doublecomplex * coef= (doublecomplex *) M.fortran_vec();
    
       zCreate_Dense_Matrix(&B, Bnr, Bnc, coef, Bnr, DN, _Z, GE);
       
-      int permc_spec = 3;
-      int perm_c[ Anc ];
-      int perm_r[ Anr ];
       oct_sparse_do_permc( permc_spec, perm_c, A );
    
       int info;
       zgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
    
-      if (info !=0 )
-         SP_FATAL_ERR("Factorization problem: zgssv");
+      if (info !=0 ) 
+	 error("sparse factorization problem: dgssv");
    
       Destroy_SuperMatrix_Store( &B );
       oct_sparse_Destroy_SuperMatrix( L ) ;
       oct_sparse_Destroy_SuperMatrix( U ) ;
+      END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    }
 
 //   return M;
@@ -1087,10 +1298,12 @@ DEFBINOP( cs_f_ldiv, complex_sparse, matrix)
 SuperMatrix assemble_sparse( int n, int m,
                              ComplexColumnVector& coefA,
                              ColumnVector& ridxA,
-                             ColumnVector& cidxA )
+                             ColumnVector& cidxA,
+                             int assemble_do_sum)
 {
    DEBUGMSG("complex_sparse - assemble_sparse");
    ASSEMBLE_SPARSE( Complex )
+// oct_sparse_verify_supermatrix( X );
    return X;
 }      
 
@@ -1114,19 +1327,29 @@ void install_complex_sparse_ops() {
    // binary operations: sparse with scalar
    //
 
-   INSTALL_BINOP (op_mul,      octave_complex_sparse, octave_scalar,         s_n_mul);
-   INSTALL_BINOP (op_mul,      octave_scalar,         octave_complex_sparse, n_s_mul);
-   INSTALL_BINOP (op_el_mul,   octave_complex_sparse, octave_scalar,         s_n_mul);
-   INSTALL_BINOP (op_el_mul,   octave_scalar,         octave_complex_sparse, n_s_mul);
-   INSTALL_BINOP (op_div,      octave_complex_sparse, octave_scalar,         s_n_div);
-   INSTALL_BINOP (op_ldiv,     octave_scalar,         octave_complex_sparse, n_s_ldiv);
+   INSTALL_BINOP (op_mul,      octave_complex_sparse, octave_scalar,         cs_n_mul);
+   INSTALL_BINOP (op_mul,      octave_scalar,         octave_complex_sparse, n_cs_mul);
+   INSTALL_BINOP (op_el_mul,   octave_complex_sparse, octave_scalar,         cs_n_mul);
+   INSTALL_BINOP (op_el_mul,   octave_scalar,         octave_complex_sparse, n_cs_mul);
+   INSTALL_BINOP (op_div,      octave_complex_sparse, octave_scalar,         cs_n_div);
+   INSTALL_BINOP (op_ldiv,     octave_scalar,         octave_complex_sparse, n_cs_ldiv);
+   INSTALL_BINOP (op_el_pow,   octave_complex_sparse, octave_scalar,         cs_n_pow);
 
-   INSTALL_BINOP (op_mul,      octave_complex_sparse, octave_complex,        s_c_mul);
-   INSTALL_BINOP (op_mul,      octave_complex,        octave_complex_sparse, c_s_mul);
-   INSTALL_BINOP (op_el_mul,   octave_complex_sparse, octave_complex,        s_c_mul);
-   INSTALL_BINOP (op_el_mul,   octave_complex,        octave_complex_sparse, c_s_mul);
-   INSTALL_BINOP (op_div,      octave_complex_sparse, octave_complex,        s_c_div);
-   INSTALL_BINOP (op_ldiv,     octave_complex,        octave_complex_sparse, c_s_ldiv);
+   INSTALL_BINOP (op_mul,      octave_complex_sparse, octave_complex,        cs_c_mul);
+   INSTALL_BINOP (op_mul,      octave_complex,        octave_complex_sparse, c_cs_mul);
+   INSTALL_BINOP (op_el_mul,   octave_complex_sparse, octave_complex,        cs_c_mul);
+   INSTALL_BINOP (op_el_mul,   octave_complex,        octave_complex_sparse, c_cs_mul);
+   INSTALL_BINOP (op_div,      octave_complex_sparse, octave_complex,        cs_c_div);
+   INSTALL_BINOP (op_ldiv,     octave_complex,        octave_complex_sparse, c_cs_ldiv);
+   INSTALL_BINOP (op_el_pow,   octave_complex_sparse, octave_complex,        cs_c_pow);
+
+   INSTALL_BINOP (op_mul,      octave_sparse,         octave_complex,        s_c_mul);
+   INSTALL_BINOP (op_mul,      octave_complex,        octave_sparse,         c_s_mul);
+   INSTALL_BINOP (op_el_mul,   octave_sparse,         octave_complex,        s_c_mul);
+   INSTALL_BINOP (op_el_mul,   octave_complex,        octave_sparse,         c_s_mul);
+   INSTALL_BINOP (op_div,      octave_sparse,         octave_complex,        s_c_div);
+   INSTALL_BINOP (op_ldiv,     octave_complex,        octave_sparse,         c_s_ldiv);
+   INSTALL_BINOP (op_el_pow,   octave_sparse,         octave_complex,        s_c_pow);
 
    //
    // binary operations: sparse with matrix 
@@ -1223,11 +1446,13 @@ LUextract(SuperMatrix *L, SuperMatrix *U, Complex *Lval, int *Lrow,
            SNptr = &((Complex*)Lstore->nzval)[L_NZ_START(j)];
 
            /* Extract U */
+	   OCTAVE_QUIT;
            for (i = U_NZ_START(j); i < U_NZ_START(j+1); ++i) {
                Uval[lastu] = ((Complex*)Ustore->nzval)[i];
                if (Uval[lastu] != 0.0) Urow[lastu++] = U_SUB(i);
            }
            /* upper triangle in the supernode */
+	   OCTAVE_QUIT;
            for (i = 0; i < upper; ++i) {
                Uval[lastu] = SNptr[i];
                if (Uval[lastu] != 0.0) Urow[lastu++] = L_SUB(istart+i);
@@ -1235,6 +1460,7 @@ LUextract(SuperMatrix *L, SuperMatrix *U, Complex *Lval, int *Lrow,
            Ucol[j+1] = lastu;
 
            /* Extract L */
+	   OCTAVE_QUIT;
            Lval[lastl] = 1.0; /* unit diagonal */
            Lrow[lastl++] = L_SUB(istart + upper - 1);
            for (i = upper; i < nsupr; ++i) {
@@ -1254,7 +1480,7 @@ LUextract(SuperMatrix *L, SuperMatrix *U, Complex *Lval, int *Lrow,
    *snnzU = lastu;
 }
 
-void
+int
 complex_sparse_LU_fact(SuperMatrix A,
                        SuperMatrix *LC,
                        SuperMatrix *UC,
@@ -1271,7 +1497,9 @@ complex_sparse_LU_fact(SuperMatrix A,
    int    info;
    int    panel_size = sp_ienv(1);
    int    relax      = sp_ienv(2);
-   int    etree[n];
+   OCTAVE_LOCAL_BUFFER (int, etree, n );
+
+   BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    SuperMatrix Ac;
    SuperMatrix L,U;
 
@@ -1283,40 +1511,46 @@ complex_sparse_LU_fact(SuperMatrix A,
 
    zgstrf(refact, &Ac, thresh, drop_tol, relax, panel_size, etree,
            NULL, 0, perm_r, perm_c, &L, &U, &info);
-   if ( info < 0 )
-      SP_FATAL_ERR ("LU factorization error");
-
-   int      snnzL, snnzU;
-
-   int       nnzL = ((SCformat*)L.Store)->nnz;
-   Complex * Lval = (Complex *) oct_sparse_malloc( nnzL * sizeof(Complex) );
-   int     * Lrow = (    int *) oct_sparse_malloc( nnzL * sizeof(    int) );
-   int     * Lcol = (    int *) oct_sparse_malloc( (n+1)* sizeof(    int) );
-
-   int       nnzU = ((NCformat*)U.Store)->nnz;
-   Complex * Uval = (Complex *) oct_sparse_malloc( nnzU * sizeof(Complex) );
-   int     * Urow = (    int *) oct_sparse_malloc( nnzU * sizeof(    int) );
-   int     * Ucol = (    int *) oct_sparse_malloc( (n+1)* sizeof(    int) );
-
-   LUextract(&L, &U, Lval, Lrow, Lcol, Uval, Urow, Ucol, &snnzL, &snnzU);
-   // we need to use the snnz values (squeezed vs. unsqueezed)
-   zCreate_CompCol_Matrix(LC, m, n, snnzL, (doublecomplex*) Lval, Lrow, Lcol, NC, _Z, GE);
-   zCreate_CompCol_Matrix(UC, m, n, snnzU, (doublecomplex*) Uval, Urow, Ucol, NC, _Z, GE);
-
-   fix_row_order_complex( *LC );
-   fix_row_order_complex( *UC );
+   if (info == 0) {
+      int      snnzL, snnzU;
+ 
+      int       nnzL = ((SCformat*)L.Store)->nnz;
+      Complex * Lval = (Complex *) oct_sparse_malloc( nnzL * sizeof(Complex) );
+      int     * Lrow = (    int *) oct_sparse_malloc( nnzL * sizeof(    int) );
+      int     * Lcol = (    int *) oct_sparse_malloc( (n+1)* sizeof(    int) );
+ 
+      int       nnzU = ((NCformat*)U.Store)->nnz;
+      Complex * Uval = (Complex *) oct_sparse_malloc( nnzU * sizeof(Complex) );
+      int     * Urow = (    int *) oct_sparse_malloc( nnzU * sizeof(    int) );
+      int     * Ucol = (    int *) oct_sparse_malloc( (n+1)* sizeof(    int) );
+ 
+      LUextract(&L, &U, Lval, Lrow, Lcol, Uval, Urow, Ucol, &snnzL, &snnzU);
+      // we need to use the snnz values (squeezed vs. unsqueezed)
+      zCreate_CompCol_Matrix(LC, m, n, snnzL, (doublecomplex*) Lval, Lrow, Lcol, NC, _Z, GE);
+      zCreate_CompCol_Matrix(UC, m, n, snnzU, (doublecomplex*) Uval, Urow, Ucol, NC, _Z, GE);
+ 
+      fix_row_order_complex( *LC );
+      fix_row_order_complex( *UC );
+   }
    
    oct_sparse_Destroy_SuperMatrix( L ) ;
    oct_sparse_Destroy_SuperMatrix( U ) ;
    oct_sparse_Destroy_SuperMatrix( Ac ) ;
    StatFree();
-
+   END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
 #if 0
    printf("verify A\n");  oct_sparse_verify_supermatrix( A );
    printf("verify LC\n"); oct_sparse_verify_supermatrix( *LC );
    printf("verify UC\n"); oct_sparse_verify_supermatrix( *UC );
 #endif   
 
+   if ( info < 0 ) {
+     error ("sparse LU factorization error");
+   } else if (info > 0) {
+     error ("sparse matrix is singlar to machine precision");
+   }
+
+   return info;
 } // complex_sparse_LU_fact(
 
 FIX_ROW_ORDER_SORT_FUNCTIONS( Complex )
@@ -1339,8 +1573,81 @@ complex_sparse_inv_uppertriang( SuperMatrix U)
    return create_SuperMatrix( Unr,Unc,cx, coefX, ridxX, cidxX );
 }                   
 
+
 /*
  * $Log: complex_sparse_ops.cc,v $
+ * Revision 1.23  2003/12/22 15:13:23  pkienzle
+ * Use error/return rather than SP_FATAL_ERROR where possible.
+ *
+ * Test for zero elements from scalar multiply/power and shrink sparse
+ * accordingly; accomodate libstdc++ bugs with mixed real/complex power.
+ *
+ * Revision 1.22  2003/12/11 22:50:29  pkienzle
+ * ... again ... remove duplicate install of ldiv (were there two?)
+ *
+ * Revision 1.21  2003/12/11 21:49:35  pkienzle
+ * Remove duplicate installation of c_cs_ldiv
+ *
+ * Revision 1.20  2003/11/23 14:21:38  adb014
+ * Octave CVS now requires void constructors for register_type
+ *
+ * Revision 1.19  2003/10/18 04:55:47  aadler
+ * spreal spimag and new tests
+ *
+ * Revision 1.18  2003/08/29 21:21:15  aadler
+ * mods to fix bugs for empty sparse
+ *
+ * Revision 1.17  2003/08/29 20:46:53  aadler
+ * fixed bug in indexing
+ *
+ * Revision 1.16  2003/08/29 19:40:56  aadler
+ * throw error rather than segfault for singular matrices
+ *
+ * Revision 1.15  2003/03/05 15:31:53  pkienzle
+ * Backport to octave-2.1.36
+ *
+ * Revision 1.14  2003/02/20 23:03:58  pkienzle
+ * Use of "T x[n]" where n is not constant is a g++ extension so replace it with
+ * OCTAVE_LOCAL_BUFFER(T,x,n), and other things to keep the picky MipsPRO CC
+ * compiler happy.
+ *
+ * Revision 1.13  2003/01/03 05:49:20  aadler
+ * mods to support 2.1.42
+ *
+ * Revision 1.12  2002/12/25 01:33:00  aadler
+ * fixed bug which allowed zero values to be stored in sparse matrices.
+ * improved print output
+ *
+ * Revision 1.11  2002/12/11 17:19:31  aadler
+ * sparse .^ scalar operations added
+ * improved test suite
+ * improved documentation
+ * new is_sparse
+ * new spabs
+ *
+ * Revision 1.10  2002/11/27 04:46:42  pkienzle
+ * Use new exception handling infrastructure.
+ *
+ * Revision 1.9  2002/11/16 20:38:20  pkienzle
+ * Windows dll and gcc 3.2 problems
+ *
+ * Revision 1.8  2002/11/05 19:21:07  aadler
+ * added indexing for complex_sparse. added tests
+ *
+ * Revision 1.7  2002/11/05 15:07:33  aadler
+ * fixed for 2.1.39 -
+ * TODO: fix complex index ops
+ *
+ * Revision 1.6  2002/11/05 06:03:25  pkienzle
+ * remove inline and parameter initialization from external function implementation
+ *
+ * Revision 1.5  2002/01/04 15:53:57  pkienzle
+ * Changes required to compile for gcc-3.0 in debian hppa/unstable
+ *
+ * Revision 1.4  2001/11/04 19:54:49  aadler
+ * fix bug with multiple entries in sparse creation.
+ * Added "summation" mode for matrix creation
+ *
  * Revision 1.3  2001/10/14 03:06:31  aadler
  * fixed memory leak in complex sparse solve
  * fixed malloc bugs for zero size allocs
