@@ -23,7 +23,7 @@ In addition to the terms of the GPL, you are permitted to link
 this program with any Open Source program, as defined by the
 Open Source Initiative (www.opensource.org)
 
-$Id: sparse_ops.cc,v 1.3 2001/10/14 03:06:31 aadler Exp $
+$Id: sparse_ops.cc,v 1.17 2003/12/22 15:13:23 pkienzle Exp $
 
 */
 
@@ -52,10 +52,12 @@ create_SuperMatrix( int nr, int nc, int nnz,
 SuperMatrix assemble_sparse( int n, int m,
                              ColumnVector& coefA,
                              ColumnVector& ridxA,
-                             ColumnVector& cidxA )
+                             ColumnVector& cidxA,
+                             int assemble_do_sum)
 {
    DEBUGMSG("sparse - assemble_sparse");
    ASSEMBLE_SPARSE( double )
+// oct_sparse_verify_supermatrix( X );
    return X;
 }      
 
@@ -64,14 +66,17 @@ SuperMatrix assemble_sparse( int n, int m,
 // Octave sparse methods
 //
 
-inline
+octave_sparse::octave_sparse (void )
+{
+      DEBUGMSG("sparse( void)");
+}
+
 octave_sparse::octave_sparse (SuperMatrix A )
 {
       DEBUGMSG("sparse( SuperMatrix A)");
       X= A;
 }
 
-inline
 octave_sparse::~octave_sparse (void)
 {
    DEBUGMSG("sparse destructor");
@@ -80,28 +85,27 @@ octave_sparse::~octave_sparse (void)
 
 //NOTE: I'm not sure when this will get called,
 //      so I don't know what to do
-inline
 octave_sparse::octave_sparse (const octave_sparse& S)
 {
    DEBUGMSG("sparse copy-constructor");
    X= S.super_matrix();
 }   
 
-inline octave_value *
-octave_sparse::clone (void)
+octave_value *
+octave_sparse::clone (void) const
 {
    DEBUGMSG("sparse - clone");
    return new octave_sparse (*this);
 }
 
-inline octave_sparse
-octave_sparse::sparse_value (bool = false) const {
+octave_sparse
+octave_sparse::sparse_value (bool) const {
    DEBUGMSG("sparse_value");
    return  (*this);
 }
 
 SuperMatrix
-octave_sparse::super_matrix (bool = false) const {
+octave_sparse::super_matrix (bool) const {
    return X;
 }
 
@@ -113,6 +117,10 @@ int octave_sparse::columns (void) const {
    return X.ncol;
 }
 
+int octave_sparse::cols (void) const {
+   return X.ncol;
+}
+
 int octave_sparse::nnz     (void) const {
    NCformat * NCF  = (NCformat * ) X.Store;
    return   NCF->nnz ;
@@ -125,6 +133,7 @@ oct_sparse_to_full ( SuperMatrix X ) {
    
    Matrix M( Xnr, Xnc );
    for (int j=0; j< Xnc; j++) {
+      OCTAVE_QUIT;
       for (int i=0; i< Xnr; i++) M(i,j)= 0;
    
       for (int i= cidxX[j]; i< cidxX[j+1]; i++) 
@@ -152,16 +161,16 @@ octave_sparse::numeric_conversion_function (void) const
 
 //idx_vector index_vector (void) const { return idx_vector ((double) iv); }
 
-octave_value octave_sparse::any (void) const {
+octave_value octave_sparse::any (int dim) const {
    DEBUGMSG("sparse - any");
    Matrix M= oct_sparse_to_full( X );
-   return M.any();
+   return M.any(dim);
 }
 
-octave_value octave_sparse::all (void) const {
+octave_value octave_sparse::all (int dim) const {
    DEBUGMSG("sparse - all");
    Matrix M= oct_sparse_to_full( X );
-   return M.all();
+   return M.all(dim);
 }
 
 bool octave_sparse::is_defined    (void) const  { return true; }
@@ -185,7 +194,7 @@ bool octave_sparse::is_true (void) const {
 // rebuild a full matrix from a sparse one
 // this functionality is accessed through 'full'
 Matrix
-octave_sparse::matrix_value (bool = false) const {
+octave_sparse::matrix_value (bool) const {
    DEBUGMSG("sparse - matrix_value");
    Matrix M= oct_sparse_to_full( X );
    return M;
@@ -225,9 +234,10 @@ oct_sparse_transpose ( SuperMatrix X ) {
 
    DECLARE_SP_POINTERS_REAL( B )
 
+   BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    dCompRow_to_CompCol( Xnc, Xnr, nnz, coefX, ridxX, cidxX,
                              &coefB, &ridxB, &cidxB);
-   
+   END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    SuperMatrix B= create_SuperMatrix( Xnc, Xnr, nnz, coefB, ridxB, cidxB );
    return B;
 }   
@@ -265,6 +275,7 @@ ixp_comp(const void *i,const void*j )
 static inline void
 sort_with_idx (sort_idx * sidx, const idx_vector& idxv, long ixl) 
 {
+   BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    if (idxv.is_colon() ) {
       for (int i=0; i< ixl; i++) {
          sidx[i].val= i;
@@ -279,6 +290,7 @@ sort_with_idx (sort_idx * sidx, const idx_vector& idxv, long ixl)
 
       qsort( sidx, ixl, sizeof(sort_idx), ixp_comp );
    }
+   END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
 }   
 
 // Return a full vector output
@@ -294,7 +306,7 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
    else  
       ixl= ix.length(-1); 
 
-   sort_idx ixp[ ixl ];
+   OCTAVE_LOCAL_BUFFER (sort_idx, ixp, ixl );
    sort_with_idx (ixp, ix, ixl);
 
    ColumnVector O( ixl );
@@ -308,11 +320,14 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
    long ip= -Xnr; // previous column position
    long jj=0,jl=0;
    for (long k=0; k< ixl; k++) {
+      OCTAVE_QUIT;
       long ii  = ixp[k].val;
       long kout= ixp[k].idx;
 
-      if ( ii<0 || ii>=Xnr*Xnc) 
-         SP_FATAL_ERR("invalid matrix index");
+      if ( ii<0 || ii>=Xnr*Xnc) {
+         error("sparse index out of range");
+	 return O;
+      }
  
       int rown= ii/Xnr;
       if ( rown > ip/Xnr ) { // we've moved to a new column
@@ -320,9 +335,9 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
          jj= cidxX[rown+1];
       }
 
-      while ( ridxX[jl] < ii%Xnr && jl < jj ) jl++;
+      while ( jl < jj && ridxX[jl] < ii%Xnr ) jl++;
 
-      if ( ridxX[jl] == ii%Xnr && jl<jj ) 
+      if ( jl<jj && ridxX[jl] == ii%Xnr ) 
          O( kout ) = coefX[jl] ;
       else
          O( kout ) = 0 ;
@@ -333,7 +348,7 @@ sparse_index_oneidx ( SuperMatrix X, const idx_vector ix) {
 } // sparse_index_oneidx (
 
 
-static SuperMatrix
+static octave_value
 sparse_index_twoidx ( SuperMatrix X,
                       const idx_vector ix,
                       const idx_vector jx) {
@@ -347,7 +362,7 @@ sparse_index_twoidx ( SuperMatrix X,
    if (jx.is_colon() )      jxl= Xnc;
    else                     jxl= jx.length(-1); 
 
-   sort_idx ixp[ ixl ];
+   OCTAVE_LOCAL_BUFFER (sort_idx, ixp, ixl );
    sort_with_idx (ixp, ix, ixl);
 
    // extimate the nnz in the output matrix
@@ -357,25 +372,34 @@ sparse_index_twoidx ( SuperMatrix X,
    int    * ridxB = intMalloc   (nnz);
    int    * cidxB = intMalloc   (jxl+1);  cidxB[0]= 0;
 
-   double tcol[ixl];  // a column of the extracted matrix
+   // a column of the extracted matrix
+   OCTAVE_LOCAL_BUFFER (double, tcol, ixl );  
 
    int cx= 0, ll=0;
-   int ip= -Xnc; // previous column position
+   // int ip= -Xnc; // previous column position
    for (int l=0; l< jxl; l++) {
+      OCTAVE_QUIT;
       if (jx.is_colon() )    ll= l;
       else                   ll= jx(l);
 
-      if ( ll<0 || ll>=Xnc) 
-            SP_FATAL_ERR("invalid matrix index (x index)");
+      if ( ll<0 || ll>=Xnc) {
+	 // XXX FIXME XXX memory leak coefB,cidxB,ridxB
+	 error("sparse column index out of range");
+	 return octave_value();
+      }
 
       int jl= cidxX[ll];
       int jj= cidxX[ll+1];
       for (long k=0; k< ixl; k++) {
+	 OCTAVE_QUIT;
          long ii  = ixp[k].val;
          long kout= ixp[k].idx;
    
-         if ( ii<0 || ii>=Xnr) 
-            SP_FATAL_ERR("invalid matrix index (x index)");
+         if ( ii<0 || ii>=Xnr) {
+	    // XXX FIXME XXX memory leak coefB,cidxB,ridxB
+            error("sparse row index out of range");
+	    return octave_value();
+	 }
 
          while ( jl < jj && ridxX[jl] < ii ) jl++;
 
@@ -385,7 +409,7 @@ sparse_index_twoidx ( SuperMatrix X,
          else
             tcol[ kout ] = 0 ;
 
-         ip=ii;
+         // ip=ii;
    
       } // for k
       for (int j=0; j<ixl; j++) {
@@ -402,12 +426,40 @@ sparse_index_twoidx ( SuperMatrix X,
    maybe_shrink( cx, nnz, ridxB, coefB );
 
    SuperMatrix B= create_SuperMatrix( ixl, jxl, cx, coefB, ridxB, cidxB );
-   return B;                          
+   return octave_value(new octave_sparse (B));
 } // sparse_index_twoidx (
 
-// indexing operations
 octave_value_list
-octave_sparse::do_multi_index_op (int, const octave_value_list& idx) 
+octave_sparse::subsref( const std::string SUBSREF_STRREF type,
+                        const LIST<octave_value_list>& idx,
+                        int nargout)
+{
+   octave_value_list retval;
+   switch (type[0]) {
+     case '(':
+       retval = do_index_op (idx.front (), 0);
+       break;
+
+     case '{':
+     case '.':
+     {
+       std::string nm = type_name ();
+       error ("%s cannot be indexed with %c", nm.c_str (), type[0]);
+     }
+     break;
+
+     default:
+       panic_impossible ();
+   }
+
+   if (idx.LISTSIZE () > 1)
+       retval = retval(0).next_subsref (type, idx);
+
+   return retval;
+}
+
+octave_value
+octave_sparse::do_index_op ( const octave_value_list& idx, int) 
 {
    DEBUGMSG("sparse - index op");
    octave_value retval;
@@ -429,10 +481,9 @@ octave_sparse::do_multi_index_op (int, const octave_value_list& idx)
       const idx_vector ix = idx (0).index_vector ();
       const idx_vector jx = idx (1).index_vector ();
 
-      retval= new octave_sparse ( 
-                  sparse_index_twoidx ( X, ix, jx ));
+      retval= sparse_index_twoidx ( X, ix, jx );
    } else
-      SP_FATAL_ERR("need 1 or 2 indices for sparse indexing operations");
+      error("need 1 or 2 indices for sparse indexing operations");
 
    return retval;
 } // octave_sparse::do_index_op
@@ -459,6 +510,7 @@ octave_sparse::extract (int r1, int c1, int r2, int c2) const {
 
    int cx= 0;
    for (int i=0, ii= c1; i < n ; i++, ii++) {
+      OCTAVE_QUIT;
       for ( int j= cidxX[ii]; j< cidxX[ii+1]; j++) {
          int row = ridxX[ j ];
          if ( row>= r1 && row<=r2 && coefX[j] !=0 ) {
@@ -479,7 +531,7 @@ octave_sparse::extract (int r1, int c1, int r2, int c2) const {
 } // octave_sparse::extract (int r1, int c1, int r2, int c2) const {
 
 void
-octave_sparse::print (ostream& os, bool pr_as_read_syntax ) const
+octave_sparse::print (std::ostream& os, bool pr_as_read_syntax ) const
 {
    DEBUGMSG("sparse - print");
 #if 0
@@ -495,9 +547,13 @@ octave_sparse::print (ostream& os, bool pr_as_read_syntax ) const
    // add one to the printed indices to go from
    //  zero-based to one-based arrays
    for (int j=0; j< Xnc; j++) 
-      for (int i= cidxX[j]; i< cidxX[j+1]; i++) 
+      for (int i= cidxX[j]; i< cidxX[j+1]; i++) {
+	 OCTAVE_QUIT;
          os << "  (" << ridxX[i]+1 <<
-               " , "  << j+1 << ") -> " << coefX[i] << "\n";
+               " , "  << j+1 << ") -> ";
+	 octave_print_internal( os, coefX[i], false );
+	 os << "\n";
+      }
 #endif                  
 } // print
 
@@ -514,20 +570,26 @@ sparse_scalar_multiply (const octave_sparse& spar,
 
   SuperMatrix X= spar.super_matrix();
   DEFINE_SP_POINTERS_REAL( X )
-  int nnz= NCFX->nnz;
+  int source_nnz= NCFX->nnz;
 
-  double *coefB = doubleMalloc(nnz);
-  int *   ridxB = intMalloc(nnz);
+  double *coefB = doubleMalloc(source_nnz);
+  int *   ridxB = intMalloc(source_nnz);
   int *   cidxB = intMalloc(X.ncol+1);
 
-  for ( int i=0; i<=Xnc; i++)
-     cidxB[i]=  cidxX[i];
-
-  for ( int i=0; i< nnz; i++) {
-     coefB[i]=  coefX[i] * s;
-     ridxB[i]=  ridxX[i];
+  int nnz=0;
+  int col=0;
+  for ( int i=0; i< source_nnz; i++) {
+     // Use volatile comparison to force zero
+     volatile double v= coefX[i] * s;
+     if (v==0.) continue;
+     coefB[nnz]= v;
+     ridxB[nnz]= ridxX[i];
+     while(i>=cidxX[col]) cidxB[col++] = nnz;
+     nnz++;
   }
 
+  while(col <= Xnc) cidxB[col++] = nnz;
+  maybe_shrink( source_nnz, nnz, ridxX, coefX );
   SuperMatrix B= create_SuperMatrix( Xnr, Xnc, nnz, coefB, ridxB, cidxB );
   return new octave_sparse ( B );
 }
@@ -558,6 +620,81 @@ DEFBINOP (n_s_ldiv, scalar, sparse) {
   return sparse_scalar_multiply (v2, 1 / d);
 }  
 
+
+DEFBINOP (s_n_pow, sparse, scalar) {
+  CAST_BINOP_ARGS (const octave_sparse&, const octave_scalar&);
+  DEBUGMSG("sparse - sparse_scalar_power");
+  double s= v2.scalar_value();
+
+  SuperMatrix X= v1.super_matrix();
+  DEFINE_SP_POINTERS_REAL( X )
+  int source_nnz= NCFX->nnz;
+
+  double *coefB = doubleMalloc(source_nnz);
+  int *   ridxB = intMalloc(source_nnz);
+  int *   cidxB = intMalloc(X.ncol+1);
+  bool power_isnot_integer= ( s != (int) s );
+
+  int nnz=0;
+  int col=0;
+  int idx=0;
+  for ( idx=0; idx< source_nnz; idx++) {
+     if (power_isnot_integer && coefX[idx]<0) break;
+     // Use volatile comparison to force zero
+     volatile double v = std::pow( coefX[idx] , s);
+     if (v==0.) continue;
+     coefB[nnz]= v;
+     ridxB[nnz]= ridxX[idx];
+     while(idx>=cidxX[col]) cidxB[col++] = nnz;
+     nnz++;
+  }
+
+  if (idx == source_nnz) { // successful processed all indices without breaking
+     while(col <= Xnc) cidxB[col++] = nnz;
+     maybe_shrink( source_nnz, nnz, ridxX, coefX );
+     SuperMatrix B= create_SuperMatrix( Xnr, Xnc, nnz, coefB, ridxB, cidxB );
+     return new octave_sparse ( B );
+  }
+  
+  // since the matrix has negative values, the output is
+  // complex. Copy what's been done to a complex, and
+  // then continue with complex output
+
+  Complex *coefBc = new_SuperLU_Complex(source_nnz);
+  for (int i=0; i< nnz; i++) {
+     coefBc[i]=  coefB[i];
+  }
+  oct_sparse_free( coefB);
+
+  for (int i=idx; i< source_nnz; i++) {
+     // Use volatile comparison to force zero
+     if (coefX[i] < 0) {
+        // Note: force both args to Complex to avoid broken libstdc++
+        coefBc[nnz] = std::pow( Complex(coefX[i],0) , Complex(s,0));
+        volatile double re = coefBc[nnz].real();
+        volatile double im = coefBc[nnz].imag();
+        if (re==0. && im==0.) continue;
+     }
+     else {
+        volatile double v = std::pow( coefX[i], s);
+	if (v==0.) continue;
+        coefBc[nnz] = v;
+     }
+     ridxB[nnz] = ridxX[i];
+     while(i>=cidxX[col]) cidxB[col++] = nnz;
+     nnz++;
+  }
+
+  while(col <= Xnc) cidxB[col++] = nnz;
+  maybe_shrink( source_nnz, nnz, ridxX, coefX );
+  SuperMatrix B= create_SuperMatrix( Xnr, Xnc, nnz, coefBc, ridxB, cidxB );
+  return new octave_complex_sparse ( B );
+}  
+
+// a scalar .^ sparse has no value, as the result will be
+// nearly full -> ie scal^0 = scal;
+//DEFBINOP (n_s_pow, scalar, sparse) {
+
 //
 // sparse by matrix  operations
 //
@@ -720,6 +857,7 @@ DEFBINOP( s_s_mul, sparse, sparse)
    int jx= 0;
    for (int j= 0 ; j < Bnc ; j++ ) {
       for (int i= 0 ; i < Anc ; i++ ) {
+	 OCTAVE_QUIT;
          int  ja= cidxA[ i ];
          int  ja_max= cidxA[ i+1 ];
          bool ja_lt_max= ja < ja_max;
@@ -733,6 +871,7 @@ DEFBINOP( s_s_mul, sparse, sparse)
          double tmpval= 0.0;
          while( ja_lt_max && jb_lt_max ) {
        
+	    OCTAVE_QUIT;
             if( ridxA_ja < ridxB_jb )
             {
                ja++; ridxA_ja= ridxA[ ja ]; ja_lt_max= ja < ja_max;
@@ -807,7 +946,7 @@ DEFBINOP( s_s_ldiv, sparse, sparse) {
    SuperMatrix   B= v2.super_matrix(); DEFINE_SP_POINTERS_REAL( B )
 // octave_value  B= new octave_sparse( v2.super_matrix() );
    int n = v1.columns();
-   int perm_c[n];
+   OCTAVE_LOCAL_BUFFER (int, perm_c, n );
    int permc_spec=3;
    octave_value_list Si= oct_sparse_inverse( v1, perm_c, permc_spec );
    octave_value inv= Si(0)*Si(1)*Si(2)*Si(3);
@@ -841,10 +980,10 @@ void oct_sparse_do_permc( int permc_spec, int perm_c[],
    int Anc= A.ncol;
 
    if ( permc_spec < 0 ) {
-      SP_FATAL_ERR("sparse solve: haven't implemented user specified permc");
-//    perm_c = perm_c_in;
+      warning("sparse solve: haven't implemented user specified permc");
+      permc_spec = 0;
    } 
-   else {
+
 //
 // KLUDGE: get_perm_c breaks (ie segfaults) if the Matrix is not
 // sufficiently sparse. It seems that this is when nnz > 0.4*(m*n)
@@ -859,12 +998,14 @@ void oct_sparse_do_permc( int permc_spec, int perm_c[],
   (double) ((NCformat *) A.Store)->nnz / (Anc*Anr) );
 #endif // VERBOSE   
 
-      if ( ((NCformat *) A.Store)->nnz >= 0.40*(Anc*Anr) )  {
-         for (int i=0 ; i< Anc ; i++)
-            perm_c[i]= i;
-      } else {
-         get_perm_c(permc_spec, &A, perm_c);
-      }
+   if ( ((NCformat *) A.Store)->nnz >= 0.40*(Anc*Anr) )  {
+      for (int i=0 ; i< Anc ; i++)
+         perm_c[i]= i;
+   } else {
+      // wrap this in the s_f_ldiv function
+      // BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
+      get_perm_c(permc_spec, &A, perm_c);
+      // END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    }
 } // static int do_permc( int permc_spec, int permc[], 
 
@@ -893,23 +1034,25 @@ DEFBINOP( s_f_ldiv, sparse, matrix)
       assert (Anc == Bnr);
       SuperMatrix L,U,B;
       double * coef= M.fortran_vec();
-   
+      int permc_spec = 3;
+      OCTAVE_LOCAL_BUFFER (int, perm_c, Anc );
+      OCTAVE_LOCAL_BUFFER (int, perm_r, Anr );
+
+      BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
       dCreate_Dense_Matrix(&B, Bnr, Bnc, coef, Bnr, DN, _D, GE);
    
-      int permc_spec = 3;
-      int perm_c[ Anc ];
-      int perm_r[ Anr ];
       oct_sparse_do_permc( permc_spec, perm_c, A );
    
       int info;
       dgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
    
       if (info !=0 )
-         SP_FATAL_ERR("Factorization problem: dgssv");
+	 error("sparse factorization problem: dgssv");
    
       Destroy_SuperMatrix_Store( &B );
       oct_sparse_Destroy_SuperMatrix( L ) ;
       oct_sparse_Destroy_SuperMatrix( U ) ;
+      END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    }
 
    return M;
@@ -938,6 +1081,7 @@ void install_sparse_ops() {
    INSTALL_BINOP (op_mul,      octave_scalar, octave_sparse, n_s_mul);
    INSTALL_BINOP (op_el_mul,   octave_sparse, octave_scalar, s_n_mul);
    INSTALL_BINOP (op_el_mul,   octave_scalar, octave_sparse, n_s_mul);
+   INSTALL_BINOP (op_el_pow,   octave_sparse, octave_scalar, s_n_pow);
 
    INSTALL_BINOP (op_div,      octave_sparse, octave_scalar, s_n_div);
    INSTALL_BINOP (op_ldiv,     octave_scalar, octave_sparse, n_s_ldiv);
@@ -1006,11 +1150,13 @@ LUextract(SuperMatrix *L, SuperMatrix *U, double *Lval, int *Lrow,
            SNptr = &((double*)Lstore->nzval)[L_NZ_START(j)];
 
            /* Extract U */
+	   OCTAVE_QUIT;
            for (i = U_NZ_START(j); i < U_NZ_START(j+1); ++i) {
                Uval[lastu] = ((double*)Ustore->nzval)[i];
                if (Uval[lastu] != 0.0) Urow[lastu++] = U_SUB(i);
            }
            /* upper triangle in the supernode */
+	   OCTAVE_QUIT;
            for (i = 0; i < upper; ++i) {
                Uval[lastu] = SNptr[i];
                if (Uval[lastu] != 0.0) Urow[lastu++] = L_SUB(istart+i);
@@ -1018,6 +1164,7 @@ LUextract(SuperMatrix *L, SuperMatrix *U, double *Lval, int *Lrow,
            Ucol[j+1] = lastu;
 
            /* Extract L */
+	   OCTAVE_QUIT;
            Lval[lastl] = 1.0; /* unit diagonal */
            Lrow[lastl++] = L_SUB(istart + upper - 1);
            for (i = upper; i < nsupr; ++i) {
@@ -1036,7 +1183,11 @@ LUextract(SuperMatrix *L, SuperMatrix *U, double *Lval, int *Lrow,
    *snnzU = lastu;
 }
 
-void
+// factors A into LC and UC, given permultations perm_c and perm_r
+// (or calculate perm_c if permc_spec indicates this)
+// return value: =0 -> success
+//               >0 -> matrix is singular, don't use for inverse solutions
+int
 sparse_LU_fact(SuperMatrix A,
                SuperMatrix *LC,
                SuperMatrix *UC,
@@ -1053,10 +1204,11 @@ sparse_LU_fact(SuperMatrix A,
    int    info;
    int    panel_size = sp_ienv(1);
    int    relax      = sp_ienv(2);
-   int    etree[n];
+   OCTAVE_LOCAL_BUFFER (int, etree, n);
    SuperMatrix Ac;
    SuperMatrix L,U;
 
+   BEGIN_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
    StatInit(panel_size, relax);
 
    oct_sparse_do_permc( permc_spec, perm_c, A);
@@ -1065,33 +1217,34 @@ sparse_LU_fact(SuperMatrix A,
 
    dgstrf(refact, &Ac, thresh, drop_tol, relax, panel_size, etree,
            NULL, 0, perm_r, perm_c, &L, &U, &info);
-   if ( info < 0 )
-      SP_FATAL_ERR ("LU factorization error");
 
-   int      snnzL, snnzU;
+   if (info == 0) {
+       int      snnzL, snnzU;
 
-   int      nnzL = ((SCformat*)L.Store)->nnz;
-   double * Lval = (double *) oct_sparse_malloc( nnzL * sizeof(double) );
-   int    * Lrow = (   int *) oct_sparse_malloc( nnzL * sizeof(   int) );
-   int    * Lcol = (   int *) oct_sparse_malloc( (n+1)* sizeof(   int) );
+       int      nnzL = ((SCformat*)L.Store)->nnz;
+       double * Lval = (double *) oct_sparse_malloc( nnzL * sizeof(double) );
+       int    * Lrow = (   int *) oct_sparse_malloc( nnzL * sizeof(   int) );
+       int    * Lcol = (   int *) oct_sparse_malloc( (n+1)* sizeof(   int) );
 
-   int      nnzU = ((NCformat*)U.Store)->nnz;
-   double * Uval = (double *) oct_sparse_malloc( nnzU * sizeof(double) );
-   int    * Urow = (   int *) oct_sparse_malloc( nnzU * sizeof(   int) );
-   int    * Ucol = (   int *) oct_sparse_malloc( (n+1)* sizeof(   int) );
+       int      nnzU = ((NCformat*)U.Store)->nnz;
+       double * Uval = (double *) oct_sparse_malloc( nnzU * sizeof(double) );
+       int    * Urow = (   int *) oct_sparse_malloc( nnzU * sizeof(   int) );
+       int    * Ucol = (   int *) oct_sparse_malloc( (n+1)* sizeof(   int) );
 
-   LUextract(&L, &U, Lval, Lrow, Lcol, Uval, Urow, Ucol, &snnzL, &snnzU);
-   // we need to use the snnz values (squeezed vs. unsqueezed)
-   dCreate_CompCol_Matrix(LC, m, n, snnzL, Lval, Lrow, Lcol, NC, _D, GE);
-   dCreate_CompCol_Matrix(UC, m, n, snnzU, Uval, Urow, Ucol, NC, _D, GE);
+       LUextract(&L, &U, Lval, Lrow, Lcol, Uval, Urow, Ucol, &snnzL, &snnzU);
+       // we need to use the snnz values (squeezed vs. unsqueezed)
+       dCreate_CompCol_Matrix(LC, m, n, snnzL, Lval, Lrow, Lcol, NC, _D, GE);
+       dCreate_CompCol_Matrix(UC, m, n, snnzU, Uval, Urow, Ucol, NC, _D, GE);
+
+       fix_row_order( *LC );
+       fix_row_order( *UC );
+   }
 
-   fix_row_order( *LC );
-   fix_row_order( *UC );
-   
    oct_sparse_Destroy_SuperMatrix( L ) ;
    oct_sparse_Destroy_SuperMatrix( U ) ;
    oct_sparse_Destroy_SuperMatrix( Ac ) ;
    StatFree();
+   END_INTERRUPT_IMMEDIATELY_IN_FOREIGN_CODE;
 
 #if 0
    printf("verify A\n");  oct_sparse_verify_supermatrix( A );
@@ -1099,6 +1252,13 @@ sparse_LU_fact(SuperMatrix A,
    printf("verify UC\n"); oct_sparse_verify_supermatrix( *UC );
 #endif   
 
+   if ( info < 0 ) {
+     error ("LU factorization error");
+   } else if (info > 0) {
+     error ("sparse matrix is singlar to machine precision");
+   }
+
+   return info;
 } // sparse_LU_fact(
 
 FIX_ROW_ORDER_SORT_FUNCTIONS( double )
@@ -1112,8 +1272,7 @@ fix_row_order( SuperMatrix X )
 }   
 
 SuperMatrix
-sparse_inv_uppertriang( SuperMatrix U)
-{
+sparse_inv_uppertriang( SuperMatrix U) {
    DEBUGMSG("sparse_inv_uppertriang");
    DEFINE_SP_POINTERS_REAL( U )
    int    nnzU= NCFU->nnz;
@@ -1123,6 +1282,60 @@ sparse_inv_uppertriang( SuperMatrix U)
 
 /*
  * $Log: sparse_ops.cc,v $
+ * Revision 1.17  2003/12/22 15:13:23  pkienzle
+ * Use error/return rather than SP_FATAL_ERROR where possible.
+ *
+ * Test for zero elements from scalar multiply/power and shrink sparse
+ * accordingly; accomodate libstdc++ bugs with mixed real/complex power.
+ *
+ * Revision 1.16  2003/11/23 14:21:39  adb014
+ * Octave CVS now requires void constructors for register_type
+ *
+ * Revision 1.15  2003/08/29 20:46:53  aadler
+ * fixed bug in indexing
+ *
+ * Revision 1.14  2003/08/29 19:40:56  aadler
+ * throw error rather than segfault for singular matrices
+ *
+ * Revision 1.13  2003/03/05 15:31:54  pkienzle
+ * Backport to octave-2.1.36
+ *
+ * Revision 1.12  2003/02/20 23:03:58  pkienzle
+ * Use of "T x[n]" where n is not constant is a g++ extension so replace it with
+ * OCTAVE_LOCAL_BUFFER(T,x,n), and other things to keep the picky MipsPRO CC
+ * compiler happy.
+ *
+ * Revision 1.11  2003/01/03 05:49:20  aadler
+ * mods to support 2.1.42
+ *
+ * Revision 1.10  2002/12/25 01:33:00  aadler
+ * fixed bug which allowed zero values to be stored in sparse matrices.
+ * improved print output
+ *
+ * Revision 1.9  2002/12/11 17:19:32  aadler
+ * sparse .^ scalar operations added
+ * improved test suite
+ * improved documentation
+ * new is_sparse
+ * new spabs
+ *
+ * Revision 1.8  2002/11/27 04:46:42  pkienzle
+ * Use new exception handling infrastructure.
+ *
+ * Revision 1.7  2002/11/05 15:07:34  aadler
+ * fixed for 2.1.39 -
+ * TODO: fix complex index ops
+ *
+ * Revision 1.6  2002/11/05 06:03:25  pkienzle
+ * remove inline and parameter initialization from external function implementation
+ *
+ * Revision 1.5  2002/01/04 15:53:57  pkienzle
+ * Changes required to compile for gcc-3.0 in debian hppa/unstable
+ *
+ * Revision 1.4  2001/11/04 19:54:49  aadler
+ * fix bug with multiple entries in sparse creation.
+ * Added "summation" mode for matrix creation
+ *
  * Revision 1.3  2001/10/14 03:06:31  aadler
  * fixed memory leak in complex sparse solve
  * fixed malloc bugs for zero size allocs
