@@ -16,9 +16,40 @@ You should have received a copy of the GNU General Public License
 along with Octave; see the file COPYING.  If not, write to the Free
 Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
-$Id: sparse_ops.h,v 1.3 2001/10/14 03:06:31 aadler Exp $
+$Id: sparse_ops.h,v 1.11 2003/12/22 15:13:23 pkienzle Exp $
 
 $Log: sparse_ops.h,v $
+Revision 1.11  2003/12/22 15:13:23  pkienzle
+Use error/return rather than SP_FATAL_ERROR where possible.
+
+Test for zero elements from scalar multiply/power and shrink sparse
+accordingly; accomodate libstdc++ bugs with mixed real/complex power.
+
+Revision 1.10  2003/10/13 03:44:26  aadler
+check for out of bounds in allocation
+
+Revision 1.9  2003/04/03 22:06:41  aadler
+sparse create bug - need to use heap for large temp vars
+
+Revision 1.8  2003/02/20 23:03:59  pkienzle
+Use of "T x[n]" where n is not constant is a g++ extension so replace it with
+OCTAVE_LOCAL_BUFFER(T,x,n), and other things to keep the picky MipsPRO CC
+compiler happy.
+
+Revision 1.7  2002/12/25 01:33:00  aadler
+fixed bug which allowed zero values to be stored in sparse matrices.
+improved print output
+
+Revision 1.6  2002/11/27 04:46:42  pkienzle
+Use new exception handling infrastructure.
+
+Revision 1.5  2002/01/04 15:53:57  pkienzle
+Changes required to compile for gcc-3.0 in debian hppa/unstable
+
+Revision 1.4  2001/11/04 19:54:49  aadler
+fix bug with multiple entries in sparse creation.
+Added "summation" mode for matrix creation
+
 Revision 1.3  2001/10/14 03:06:31  aadler
 fixed memory leak in complex sparse solve
 fixed malloc bugs for zero size allocs
@@ -37,7 +68,6 @@ Revision 1.1  2001/03/30 04:34:23  aadler
 
 */
 
-
 // I would like to do this with templates,
 // but I don't think you can specify operators
 //
@@ -68,6 +98,7 @@ Revision 1.1  2001/03/30 04:34:23  aadler
          bool jb_lt_max= jb < jb_max; \
  \
          while( ja_lt_max || jb_lt_max ) { \
+            OCTAVE_QUIT; \
             if( ( !jb_lt_max ) || \
                 ( ja_lt_max && (ridxA[ja] < ridxB[jb]) ) ) \
             { \
@@ -121,6 +152,7 @@ Revision 1.1  2001/03/30 04:34:23  aadler
       for (int i=0; i < Anc ; i++) { \
          for (int j= cidxA[i]  ; \
                   j< cidxA[i+1]; j++) { \
+            OCTAVE_QUIT; \
             int  rowidx = ridxA[j]; \
             TYPX tmpval = B(rowidx,i); \
             if (tmpval != 0.0) { \
@@ -146,9 +178,10 @@ Revision 1.1  2001/03/30 04:34:23  aadler
       assert (Anc == Bnr); \
       for (int i=0; i< Anr; i++ ) \
          for (int j=0; j< Bnc; j++ ) \
-            X.elem(i,j)=0; \
+            X.elem(i,j)=0.; \
       for ( int i=0; i < Anc ; i++) { \
          for ( int j= cidxA[i]; j< cidxA[i+1]; j++) { \
+            OCTAVE_QUIT; \
             int  col = ridxA[j]; \
             TYPX tmpval = coefA[j]; \
             for ( int k=0 ; k< Bnc; k++) { \
@@ -166,9 +199,10 @@ Revision 1.1  2001/03/30 04:34:23  aadler
       assert (Anc == Bnr); \
       for (int i=0; i< Anr; i++ ) \
          for (int j=0; j< Bnc; j++ ) \
-            X.elem(i,j)=0; \
+            X.elem(i,j)=0.; \
       for ( int i=0; i < Bnc ; i++) { \
          for ( int j= cidxB[i]; j< cidxB[i+1]; j++) { \
+            OCTAVE_QUIT; \
             int  col = ridxB[j]; \
             TYPX tmpval = coefB[j]; \
             for ( int k=0 ; k< Anr; k++) { \
@@ -204,20 +238,22 @@ Revision 1.1  2001/03/30 04:34:23  aadler
       TYPX * Xcol= (TYPX*)oct_sparse_malloc( Anr  * sizeof(TYPX)); \
       int cx= 0; \
       for ( int i=0; i < Bnc ; i++) { \
-         for (int k=0; k<Anr; k++) Xcol[k]= 0; \
+         OCTAVE_QUIT; \
+         for (int k=0; k<Anr; k++) Xcol[k]= 0.; \
          for (int j= cidxB[i]; j< cidxB[i+1]; j++) { \
             int  col= ridxB[j]; \
             TYPX tmpval = coefB[j]; \
             for (int k= cidxA[col] ; k< cidxA[col+1]; k++)  \
                Xcol[ ridxA[k] ]+= tmpval * coefA[k]; \
          } \
-         for (int k=0; k<Anr; k++)  \
-            if ( Xcol[k] !=0 ) { \
+         for (int k=0; k<Anr; k++) { \
+            if ( Xcol[k] !=0. ) { \
                check_bounds( cx, nnz, ridxX, coefX ); \
                ridxX[ cx ]= k; \
                coefX[ cx ]= Xcol[k]; \
                cx++; \
             } \
+         } \
          cidxX[i+1] = cx; \
       } \
       free( Xcol ); \
@@ -231,6 +267,10 @@ Revision 1.1  2001/03/30 04:34:23  aadler
 //
 // NOTE2: be careful about when we convert ri to int,
 // otherwise the maximum matrix size will be m*n < maxint/2
+// 
+// The OCTAVE_LOCAL_BUFFER can't be used for the sort index
+//  when the requested size is too big. we need to put it on the
+//  heap.
 #define ASSEMBLE_SPARSE( TYPX ) \
    int  nnz= MAX( ridxA.length(), cidxA.length() ); \
    TYPX* coefX= (TYPX*)oct_sparse_malloc( nnz  * sizeof(TYPX)); \
@@ -241,28 +281,58 @@ Revision 1.1  2001/03/30 04:34:23  aadler
    bool ci_scalar = (cidxA.length() == 1); \
    bool cf_scalar = (coefA.length() == 1); \
  \
-   sort_idxl idx[ nnz ]; \
+   sort_idxl* sidx = (sort_idxl *)oct_sparse_malloc( nnz* sizeof(sort_idxl) );\
+   int actual_nnz=0; \
+   OCTAVE_QUIT; \
    for (int i=0; i<nnz; i++) { \
-      idx[i].val = (unsigned long) ( \
-                ( ri_scalar ? ridxA(0) : ridxA(i) ) - 1 + \
-           m * (( ci_scalar ? cidxA(0) : cidxA(i) ) - 1) );  \
-      idx[i].idx = i; \
+      double rowidx=  ( ri_scalar ? ridxA(0) : ridxA(i) ) - 1; \
+      double colidx=  ( ci_scalar ? cidxA(0) : cidxA(i) ) - 1;  \
+      if (rowidx >= m || rowidx < 0) \
+         error("sparse row index out of range"); \
+      else if (colidx >= n || colidx < 0) \
+         error("sparse column index out of range"); \
+      if (error_state) { actual_nnz=0; break; } \
+      if ( coefA(cf_scalar?0:i) != 0. ) { \
+         sidx[actual_nnz].val = (long) ( rowidx + m*colidx ); \
+         sidx[actual_nnz].idx = i; \
+         actual_nnz++; \
+      } \
    } \
  \
-   qsort( idx, nnz, sizeof(sort_idxl), sidxl_comp ); \
+   OCTAVE_QUIT; \
+   qsort( sidx, actual_nnz, sizeof(sort_idxl), sidxl_comp ); \
     \
+   OCTAVE_QUIT; \
    int cx= 0; \
-   for (int i=0; i<nnz; i++) { \
-      unsigned long ii= (int) idx[i].idx; \
-      coefX[i]=      ( cf_scalar ? coefA(0) : coefA(ii) ); \
-      double ri  =   ( ri_scalar ? ridxA(0) : ridxA(ii) ) - 1 ; \
-      ridxX[i]= (int) (ri - ((int) (ri/m))*m ) ; \
-      int ci  = (int)( ci_scalar ? cidxA(0) : cidxA(ii) ) - 1 ; \
-      while( cx < ci ) cidxX[++cx]= i; \
+   long prev_val=-1;  \
+   int ii= -1; \
+   for (int i=0; i<actual_nnz; i++) { \
+      OCTAVE_QUIT; \
+      long  idx= (long) sidx[i].idx; \
+      long  val= (long) sidx[i].val; \
+      if (prev_val < val) { \
+         ii++; \
+         coefX[ii]=     ( cf_scalar ? coefA(0) : coefA(idx) ); \
+         double ri  =   ( ri_scalar ? ridxA(0) : ridxA(idx) ) - 1 ; \
+         ridxX[ii]= (long) (ri - ((long) (ri/m))*m ) ; \
+         long ci  = (long)( ci_scalar ? cidxA(0) : cidxA(idx) ) - 1 ; \
+         while( cx < ci ) cidxX[++cx]= ii; \
+      } else { \
+         if (assemble_do_sum) \
+            coefX[ii]+=      ( cf_scalar ? coefA(0) : coefA(idx) ); \
+         else \
+            coefX[ii] =      ( cf_scalar ? coefA(0) : coefA(idx) ); \
+      } \
+      prev_val= val;\
    } \
-   while( cx < n ) cidxX[++cx]= nnz; \
+   while( cx < n ) cidxX[++cx]= ii+1; \
+   OCTAVE_QUIT; \
+   maybe_shrink( ii+1, actual_nnz, ridxX, coefX ); \
  \
-   SuperMatrix X= create_SuperMatrix( m, n, nnz, coefX, ridxX, cidxX );
+   OCTAVE_QUIT; \
+   SuperMatrix X= create_SuperMatrix( m, n, ii+1, coefX, ridxX, cidxX ); \
+   oct_sparse_free( sidx );
+
 
 // assemble a sparse matrix from full
 //   called by one arg for sparse
@@ -276,8 +346,9 @@ Revision 1.1  2001/03/30 04:34:23  aadler
    int jx=0; \
    for (int j= 0; j<Anc ; j++ ) { \
       for (int i= 0; i<Anr ; i++ ) { \
+         OCTAVE_QUIT; \
          TYPX tmpval= A(i,j); \
-         if (tmpval != 0) { \
+         if (tmpval != 0.) { \
             check_bounds( jx, nnz, ridx, coef ); \
             ridx[ jx ]= i; \
             coef[ jx ]= tmpval; \
@@ -325,25 +396,25 @@ fixrow_comp( const void *i, const void *j)  \
 // }
 
 #define FIX_ROW_ORDER_FUNCTIONS \
-   int    nnz= NCFX->nnz; \
    for ( int i=0; i < Xnr ; i++) { \
-      assert( cidxX[i] >= 0); \
-      assert( cidxX[i] <  nnz); \
+      OCTAVE_QUIT; \
+      assert( cidxX[i] >= 0.); \
+      assert( cidxX[i] <= NCFX->nnz); \
       assert( cidxX[i] <=  cidxX[i+1]); \
       int reorder=0; \
       for( int j= cidxX[i]; \
                j< cidxX[i+1]; \
                j++ ) { \
-         assert( ridxX[j] >= 0); \
+         assert( ridxX[j] >= 0.); \
          assert( ridxX[j] <  Xnc); \
-         assert( coefX[j] !=  0); /* don't keep zero values */ \
+         assert( coefX[j] !=  0.); /* don't keep zero values */ \
          if (j> cidxX[i]) \
             if ( ridxX[j-1] > ridxX[j] ) \
                reorder=1; \
       } /* for j */ \
       if(reorder) { \
          int snum= cidxX[i+1] - cidxX[i]; \
-         fixrow_sort arry[snum]; \
+         OCTAVE_LOCAL_BUFFER (fixrow_sort, arry, snum); \
          /* copy to the sort struct */ \
          for( int k=0, \
                   j= cidxX[i]; \
@@ -353,7 +424,9 @@ fixrow_comp( const void *i, const void *j)  \
             arry[k].val= coefX[j]; \
             k++; \
          } \
+         OCTAVE_QUIT; \
          qsort( arry, snum, sizeof(fixrow_sort), fixrow_comp ); \
+         OCTAVE_QUIT; \
          /* copy back to the position */ \
          for( int k=0, \
                   j= cidxX[i]; \
@@ -422,6 +495,7 @@ fixrow_comp( const void *i, const void *j)  \
  \
    /* iterate accross columns of output matrix */ \
    for ( int n=0; n < Unr ; n++) { \
+      OCTAVE_QUIT; \
       /* place the 1 in the identity position */ \
       int cx_colstart= cx; \
  \
@@ -433,13 +507,14 @@ fixrow_comp( const void *i, const void *j)  \
  \
       /* iterate accross columns of input matrix */ \
       for ( int m= n+1; m< Unr; m++) { \
-         TYPE v=0; \
+         TYPE v=0.; \
          /* iterate to calculate sum */ \
          int colXp= cidxX[n]; \
          int colUp= cidxU[m]; \
  \
          int rpX, rpU; \
          do { \
+            OCTAVE_QUIT; \
             rpX= ridxX [ colXp ]; \
             rpU= ridxU [ colUp ]; \
  \
@@ -464,9 +539,9 @@ fixrow_comp( const void *i, const void *j)  \
          assert (ridxU[ colUp ] == m ); /* assert  U is upper triangular */ \
  \
          TYPE pivot= coefU[ colUp ]; \
-         if (pivot == 0) gripe_divide_by_zero (); \
+         if (pivot == 0.) gripe_divide_by_zero (); \
  \
-         if (v!=0) { \
+         if (v!=0.) { \
             check_bounds( cx, nnz, ridxX, coefX ); \
             ridxX[cx]= m; \
             coefX[cx]= v / pivot; \
@@ -478,7 +553,7 @@ fixrow_comp( const void *i, const void *j)  \
       int colUp= cidxU[n+1]-1; \
       assert (ridxU[ colUp ] == n ); /* assert U upper triangular */ \
       TYPE pivot= coefU[ colUp ]; \
-      if (pivot == 0) gripe_divide_by_zero (); \
+      if (pivot == 0.) gripe_divide_by_zero (); \
  \
       if (pivot!=1.0) \
          for( int i= cx_colstart; i< cx; i++)  \
